from audiomate.corpus.io import CorpusDownloader, available_downloaders, create_downloader_of_type
from audiomate.corpus.io import CorpusReader, available_readers, create_reader_of_type
from audiomate.corpus.io import CorpusWriter, available_writers, create_writer_of_type
from audiomate.corpus.io import UnknownDownloaderException, UnknownReaderException, UnknownWriterException
from audiomate.corpus.io.downloader import ArchiveDownloader

import pytest


def test_all_downloaders_registered():
    expected_downloaders = CorpusDownloader.__subclasses__()
    expected_downloaders.remove(ArchiveDownloader)
    actual_downloaders = available_downloaders()

    assert len(expected_downloaders) == len(actual_downloaders), \
        'Number of registered downloaders does not match number of present downloaders'

    for expected_downloader in expected_downloaders:
        assert expected_downloader in actual_downloaders.values(), \
            'Downloader not registered'
        assert expected_downloader.type() in actual_downloaders.keys(), \
            'Downloader not available under its type()'


def test_all_downloaders_creatable():
    expected_downloaders = CorpusDownloader.__subclasses__()
    expected_downloaders.remove(ArchiveDownloader)

    for expected_downloader in expected_downloaders:
        assert isinstance(
            create_downloader_of_type(expected_downloader.type()),
            expected_downloader
        )


def test_unknown_downloader_creation_throws():
    with pytest.raises(UnknownDownloaderException, message='Unknown downloader: does_not_exist'):
        create_downloader_of_type('does_not_exist')


def test_all_readers_registered():
    expected_readers = CorpusReader.__subclasses__()
    actual_readers = available_readers()

    assert len(expected_readers) == len(actual_readers), \
        'Number of registered readers does not match number of present readers'

    for expected_reader in expected_readers:
        assert expected_reader in actual_readers.values(), \
            'Reader not registered'
        assert expected_reader.type() in actual_readers.keys(), \
            'Reader not available under its type()'


def test_all_readers_creatable():
    expected_readers = CorpusReader.__subclasses__()

    for expected_reader in expected_readers:
        assert isinstance(create_reader_of_type(expected_reader.type()), expected_reader)


def test_unknown_reader_creation_throws():
    with pytest.raises(UnknownReaderException, message='Unknown reader: does_not_exist'):
        create_reader_of_type('does_not_exist')


def test_all_writers_registered():
    expected_writers = CorpusWriter.__subclasses__()
    actual_writers = available_writers()

    assert len(expected_writers) == len(actual_writers), \
        'Number of registered writers does not match number of present writers'

    for expected_writer in expected_writers:
        assert expected_writer in actual_writers.values(), \
            'Writer not registered'
        assert expected_writer.type() in actual_writers.keys(), \
            'Writer not available under its type()'


def test_all_writers_creatable():
    expected_writers = CorpusWriter.__subclasses__()

    for expected_writer in expected_writers:
        assert isinstance(create_writer_of_type(expected_writer.type()), expected_writer)


def test_unknown_writer_creation_throws():
    with pytest.raises(UnknownWriterException, message='Unknown writer: does_not_exist'):
        create_writer_of_type('does_not_exist')



import os

import pytest
import numpy as np
import librosa

from audiomate import tracks

from tests import resources


@pytest.fixture()
def audio_path():
    return os.path.join(os.path.dirname(resources.__file__), 'audio_formats')


class TestFile:

    @pytest.mark.parametrize('name,sampling_rate', [
        ('flac_1_16k_16b.flac', 16000),
        ('mp3_2_44_1k_16b.mp3', 44100),
        ('wav_1_16k_24b.wav', 16000),
        ('wav_2_44_1k_16b.wav', 44100),
        ('wavex_2_48k_24b.wav', 48000)
    ])
    def test_sampling_rate(self, name, sampling_rate, audio_path):
        file_obj = tracks.FileTrack('some_idx', os.path.join(audio_path, name))

        assert file_obj.sampling_rate == sampling_rate

    @pytest.mark.parametrize('name,num_channels', [
        ('flac_1_16k_16b.flac', 1),
        ('mp3_2_44_1k_16b.mp3', 2),
        ('wav_1_16k_24b.wav', 1),
        ('wav_2_44_1k_16b.wav', 2),
        ('wavex_2_48k_24b.wav', 2)
    ])
    def test_num_channels(self, name, num_channels, audio_path):
        file_obj = tracks.FileTrack('some_idx', os.path.join(audio_path, name))

        assert file_obj.num_channels == num_channels

    @pytest.mark.parametrize('name,num_samples', [
        ('wav_1_16k_24b.wav', 41523),
        ('wav_2_44_1k_16b.wav', 176400),
        ('wavex_2_48k_24b.wav', 192000)
    ])
    def test_num_samples(self, name, num_samples, audio_path):
        file_obj = tracks.FileTrack('some_idx', os.path.join(audio_path, name))

        assert file_obj.num_samples == num_samples

    @pytest.mark.parametrize('name,num_samples', [
        ('flac_1_16k_16b.flac', 103424),
        ('mp3_2_44_1k_16b.mp3', 222336)
    ])
    def test_num_samples_compressed_formats(self, name, num_samples, audio_path):
        file_obj = tracks.FileTrack('some_idx', os.path.join(audio_path, name))

        assert file_obj.num_samples == pytest.approx(num_samples, abs=2000)

    @pytest.mark.parametrize('name,duration', [
        ('wav_1_16k_24b.wav', 2.5951875),
        ('wav_2_44_1k_16b.wav', 4.0),
        ('wavex_2_48k_24b.wav', 4.0)
    ])
    def test_num_duration(self, name, duration, audio_path):
        file_obj = tracks.FileTrack('some_idx', os.path.join(audio_path, name))

        assert file_obj.duration == pytest.approx(duration)

    @pytest.mark.parametrize('name,duration', [
        ('flac_1_16k_16b.flac', 6.464),
        ('mp3_2_44_1k_16b.mp3', 5.0416326531)
    ])
    def test_num_duration_compressed_formats(self, name, duration, audio_path):
        file_obj = tracks.FileTrack('some_idx', os.path.join(audio_path, name))

        assert file_obj.duration == pytest.approx(duration, abs=0.1)

    @pytest.mark.parametrize('name', [
        ('flac_1_16k_16b.flac'),
        ('mp3_2_44_1k_16b.mp3'),
        ('wav_1_16k_24b.wav'),
        ('wav_2_44_1k_16b.wav'),
        ('wavex_2_48k_24b.wav')
    ])
    def test_read_samples(self, name, audio_path):
        audio_path = os.path.join(audio_path, name)
        file_obj = tracks.FileTrack('some_idx', audio_path)

        expected, __ = librosa.core.load(audio_path, sr=None, mono=True)
        actual = file_obj.read_samples()

        assert np.array_equal(actual, expected)

    @pytest.mark.parametrize('name', [
        ('flac_1_16k_16b.flac'),
        ('mp3_2_44_1k_16b.mp3'),
        ('wav_1_16k_24b.wav'),
        ('wav_2_44_1k_16b.wav'),
        ('wavex_2_48k_24b.wav')
    ])
    def test_read_samples_fix_sampling_rate(self, name, audio_path):
        audio_path = os.path.join(audio_path, name)
        file_obj = tracks.FileTrack('some_idx', audio_path)

        expected, __ = librosa.core.load(audio_path, sr=16000, mono=True)
        actual = file_obj.read_samples(sr=16000)

        assert np.array_equal(actual, expected)

    @pytest.mark.parametrize('name', [
        ('flac_1_16k_16b.flac'),
        ('mp3_2_44_1k_16b.mp3'),
        ('wav_1_16k_24b.wav'),
        ('wav_2_44_1k_16b.wav'),
        ('wavex_2_48k_24b.wav')
    ])
    def test_read_samples_range(self, name, audio_path):
        audio_path = os.path.join(audio_path, name)
        file_obj = tracks.FileTrack('some_idx', audio_path)

        expected, __ = librosa.core.load(audio_path, sr=None, mono=True,
                                         offset=1.0, duration=1.7)
        actual = file_obj.read_samples(offset=1.0, duration=1.7)

        assert np.array_equal(actual, expected)

    def test_read_frames(self, tmpdir):
        wav_path = os.path.join(tmpdir.strpath, 'file.wav')
        wav_content = np.random.random(10044)

        librosa.output.write_wav(wav_path, wav_content, 16000)
        file_obj = tracks.FileTrack('some_idx', wav_path)

        data = list(file_obj.read_frames(frame_size=400, hop_size=160))
        frames = np.array([x[0] for x in data])
        last = [x[1] for x in data]

        assert frames.shape == (62, 400)
        assert frames.dtype == np.float32
        assert np.allclose(frames[0], wav_content[:400], atol=0.0001)
        expect = np.pad(wav_content[9760:], (0, 116), mode='constant')
        assert np.allclose(frames[61], expect, atol=0.0001)

        assert last[:-1] == [False] * (len(data) - 1)
        assert last[-1]



import os

from . import base
import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus import subset
from audiomate.utils import textfile


class FluentSpeechReader(base.CorpusReader):
    """
    Reader for the Fluent Speech Commands Dataset.

    .. seealso::

       `Fluent Speech Commands Dataset <http://www.fluent.ai/research/fluent-speech-commands/>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'fluent-speech'

    def _check_for_missing_files(self, path):
        files = [
            os.path.join(path, 'data', 'speaker_demographics.csv'),
            os.path.join(path, 'data', 'train_data.csv'),
            os.path.join(path, 'data', 'test_data.csv'),
            os.path.join(path, 'data', 'valid_data.csv'),
        ]

        missing = []

        for path in files:
            if not os.path.isfile(path):
                missing.append(path)

        return missing

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        speaker_path = os.path.join(path, 'data', 'speaker_demographics.csv')
        speakers = FluentSpeechReader.load_speakers(speaker_path)

        FluentSpeechReader.load_part(path, 'train', corpus, speakers)
        FluentSpeechReader.load_part(path, 'valid', corpus, speakers)
        FluentSpeechReader.load_part(path, 'test', corpus, speakers)

        return corpus

    @staticmethod
    def load_part(base_path, part_name, corpus, speakers):
        part_file_path = os.path.join(base_path, 'data', '{}_data.csv'.format(part_name))
        entries = textfile.read_separated_lines_generator(
            part_file_path,
            separator=',',
            max_columns=7,
            ignore_lines_starting_with=[',']
        )

        part_ids = []

        for entry in entries:
            file_path = entry[1]
            file_base = os.path.basename(file_path)
            idx = os.path.splitext(file_base)[0]
            speaker_idx = entry[2]
            part_ids.append(idx)

            if speaker_idx not in corpus.issuers.keys():
                corpus.import_issuers(speakers[speaker_idx])

            track = corpus.new_file(
                os.path.join(base_path, file_path),
                idx
            )

            utt = corpus.new_utterance(
                idx,
                track.idx,
                speaker_idx
            )

            transcription = annotations.LabelList.create_single(
                entry[3],
                idx=audiomate.corpus.LL_WORD_TRANSCRIPT
            )
            utt.set_label_list(transcription)

            if entry[4] != 'none':
                action = annotations.LabelList.create_single(
                    entry[4],
                    idx='action'
                )
                utt.set_label_list(action)

            if entry[5] != 'none':
                object_label = annotations.LabelList.create_single(
                    entry[5],
                    idx='object'
                )
                utt.set_label_list(object_label)

            if entry[6] != 'none':
                location = annotations.LabelList.create_single(
                    entry[6],
                    idx='location'
                )
                utt.set_label_list(location)

        filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=set(part_ids))
        subview = subset.Subview(corpus, filter_criteria=[filter])
        corpus.import_subview(part_name, subview)

    @staticmethod
    def load_speakers(path):
        entries = textfile.read_separated_lines_generator(
            path,
            separator=',',
            max_columns=6,
            ignore_lines_starting_with=['speakerId']
        )

        idx_to_speaker = {}

        for entry in entries:
            spk = FluentSpeechReader.parse_speaker_record(entry)
            idx_to_speaker[spk.idx] = spk

        return idx_to_speaker

    @staticmethod
    def parse_speaker_record(record):
        idx = record[0]

        gender = issuers.Gender.UNKNOWN

        if record[4] == 'male':
            gender = issuers.Gender.MALE
        elif record[4] == 'female':
            gender = issuers.Gender.FEMALE

        age_group = issuers.AgeGroup.UNKNOWN

        if record[5] in ('22-40', '41-65'):
            age_group = issuers.AgeGroup.ADULT
        elif record[5] == '65+':
            age_group = issuers.AgeGroup.SENIOR

        native_lang = None

        if record[2].startswith('English'):
            native_lang = 'eng'
        elif record[2].startswith('French'):
            native_lang = 'fra'
        elif record[2].startswith('Spanish'):
            native_lang = 'spa'
        elif record[2].startswith('Telugu'):
            native_lang = 'tel'

        return issuers.Speaker(idx, gender, age_group, native_lang)



"""
The :mod:`audiomate.feeding` module provides tools for a simple access to data stored in different
:class:`audiomate.corpus.assets.Container`.
"""

from .dataset import Dataset  # noqa: F401
from .dataset import UtteranceDataset  # noqa: F401
from .dataset import FrameDataset  # noqa: F401
from .dataset import MultiFrameDataset  # noqa: F401

from .iterator import DataIterator  # noqa: F401
from .iterator import FrameIterator  # noqa: F401
from .iterator import MultiFrameIterator  # noqa: F401

from .partitioning import PartitioningFeatureIterator  # noqa: F401

from .partitioning import PartitioningContainerLoader  # noqa: F401
from .partitioning import PartitionInfo  # noqa: F401
from .partitioning import PartitionData  # noqa: F401



import numpy as np

from . import base


class TokenOrdinalEncoder(base.Encoder):
    """
    Class to encode labels of a given label-list. Every token of the labels is mapped to a number.
    For the full utterance a sequence/array of numbers are computed, which correspond to tokens.

    Tokens are extracted from labels by splitting using a delimiter (by default space).
    See :meth:`audiomate.annotations.Label.tokenized`.
    Hence a token can be word, phone, ..., depending on the label and the delimiter.

    Args:
        label_list_idx (str): The name of the label-list to use for encoding.
                              Only labels of this label-list are considered.
        tokens (list): List of tokens that defines the mapping. First label will get the 0 in the encoding and so on.
        token_delimiter (str): Delimiter to split labels into tokens.

    Example:

        >>> ll = LabelList(idx='words', labels=[Label('down the  road')])
        >>> utt = Utterance('utt-1', 'file-x', label_lists=ll)
        >>>
        >>> tokens = ['up', 'down', 'road', 'stree', 'the']
        >>> encoder = TokenOrdinalEncoder('words', tokens, token_delimiter=' ')
        >>> encoder.encode_utterance(utt)
        np.array([1, 4, 2])
    """

    def __init__(self, label_list_idx, tokens, token_delimiter=' '):
        self.label_list_idx = label_list_idx
        self.tokens = tokens
        self.token_delimiter = token_delimiter

    def encode_utterance(self, utterance, corpus=None):
        if self.label_list_idx not in utterance.label_lists.keys():
            raise ValueError('Utterance {} has not label-list with idx {}!'.format(utterance.idx, self.label_list_idx))

        ll = utterance.label_lists[self.label_list_idx]
        concatenated_tokens = ll.tokenized(delimiter=self.token_delimiter, overlap_threshold=0.1)

        ordinals = []

        for t in concatenated_tokens:
            if t in self.tokens:
                ordinals.append(self.tokens.index(t))
            else:
                raise ValueError('Token {} not in token-list!'.format(t))

        return np.array(ordinals)



import os

import numpy as np
import h5py

from audiomate.feeding import PartitioningFeatureIterator
from audiomate.feeding import partitioning
from audiomate import containers

import pytest


class TestPartitioningContainerLoader:

    def test_scan_computes_correct_size_for_one_container(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c1.open()
        c1.set('utt-1', np.random.random((6, 6)).astype(np.float32))
        c1.set('utt-2', np.random.random((2, 6)).astype(np.float32))
        c1.set('utt-3', np.random.random((9, 6)).astype(np.float32))

        loader = partitioning.PartitioningContainerLoader(['utt-1', 'utt-2', 'utt-3'],
                                                          c1, '250', shuffle=True, seed=88)

        sizes = loader._scan()

        assert sizes == {
            'utt-1': 6 * 6 * np.dtype(np.float32).itemsize,
            'utt-2': 2 * 6 * np.dtype(np.float32).itemsize,
            'utt-3': 9 * 6 * np.dtype(np.float32).itemsize
        }

    def test_scan_computes_correct_size_for_multiple_containers(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c2 = containers.Container(os.path.join(tmpdir.strpath, 'c2.h5'))
        c3 = containers.Container(os.path.join(tmpdir.strpath, 'c3.h5'))
        c1.open()
        c1.set('utt-1', np.random.random((6, 6)).astype(np.float32))
        c1.set('utt-2', np.random.random((2, 6)).astype(np.float32))
        c1.set('utt-3', np.random.random((9, 6)).astype(np.float32))
        c2.open()
        c2.set('utt-1', np.random.random((2, 6)).astype(np.float32))
        c2.set('utt-2', np.random.random((1, 6)).astype(np.float32))
        c2.set('utt-3', np.random.random((4, 6)).astype(np.float32))
        c3.open()
        c3.set('utt-1', np.random.random((1, 6)).astype(np.float32))
        c3.set('utt-2', np.random.random((3, 6)).astype(np.float32))
        c3.set('utt-3', np.random.random((8, 6)).astype(np.float32))

        loader = partitioning.PartitioningContainerLoader(['utt-1', 'utt-2', 'utt-3'],
                                                          [c1, c2, c3], '1000', shuffle=True, seed=88)

        sizes = loader._scan()

        assert sizes == {
            'utt-1': (6 + 2 + 1) * 6 * np.dtype(np.float32).itemsize,
            'utt-2': (2 + 1 + 3) * 6 * np.dtype(np.float32).itemsize,
            'utt-3': (9 + 4 + 8) * 6 * np.dtype(np.float32).itemsize
        }

    def test_get_lengths_returns_correct_lengths_for_multiple_containers(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c2 = containers.Container(os.path.join(tmpdir.strpath, 'c2.h5'))
        c3 = containers.Container(os.path.join(tmpdir.strpath, 'c3.h5'))
        c1.open()
        c1.set('utt-1', np.random.random((6, 6)).astype(np.float32))
        c1.set('utt-2', np.random.random((2, 6)).astype(np.float32))
        c1.set('utt-3', np.random.random((9, 6)).astype(np.float32))
        c2.open()
        c2.set('utt-1', np.random.random((2, 6)).astype(np.float32))
        c2.set('utt-2', np.random.random((1, 6)).astype(np.float32))
        c2.set('utt-3', np.random.random((4, 6)).astype(np.float32))
        c3.open()
        c3.set('utt-1', np.random.random((1, 6)).astype(np.float32))
        c3.set('utt-2', np.random.random((3, 6)).astype(np.float32))
        c3.set('utt-3', np.random.random((8, 6)).astype(np.float32))

        loader = partitioning.PartitioningContainerLoader(['utt-1', 'utt-2', 'utt-3'],
                                                          [c1, c2, c3], '1000', shuffle=True, seed=88)

        lengths = loader._get_all_lengths()

        assert len(lengths) == 3
        assert lengths['utt-1'] == (6, 2, 1)
        assert lengths['utt-2'] == (2, 1, 3)
        assert lengths['utt-3'] == (9, 4, 8)

    def test_raises_error_if_utt_is_missing_in_container(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c1.open()
        c1.set('utt-1', np.random.random((6, 6)).astype(np.float32))
        c1.set('utt-3', np.random.random((9, 6)).astype(np.float32))

        with pytest.raises(ValueError):
            partitioning.PartitioningContainerLoader(['utt-1', 'utt-2', 'utt-3'],
                                                     c1, '250', shuffle=True, seed=88)

    def test_raises_error_if_utt_is_to_large_for_partition_size(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c1.open()
        c1.set('utt-1', np.random.random((6, 6)).astype(np.float32))
        # Needs 264
        c1.set('utt-3', np.random.random((11, 6)).astype(np.float32))

        with pytest.raises(ValueError):
            partitioning.PartitioningContainerLoader(['utt-1', 'utt-3'],
                                                     c1, '250', shuffle=True, seed=88)

    def test_reload_creates_correct_partitions(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c1.open()
        c1.set('utt-1', np.random.random((6, 6)).astype(np.float32))
        c1.set('utt-2', np.random.random((2, 6)).astype(np.float32))
        c1.set('utt-3', np.random.random((9, 6)).astype(np.float32))
        c1.set('utt-4', np.random.random((2, 6)).astype(np.float32))
        c1.set('utt-5', np.random.random((5, 6)).astype(np.float32))

        loader = partitioning.PartitioningContainerLoader(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                                          c1, '250', shuffle=False)

        assert len(loader.partitions) == 3
        assert loader.partitions[0].utt_ids == ['utt-1', 'utt-2']
        assert loader.partitions[0].utt_lengths == [(6,), (2,)]
        assert loader.partitions[0].size == 192
        assert loader.partitions[1].utt_ids == ['utt-3']
        assert loader.partitions[1].utt_lengths == [(9,)]
        assert loader.partitions[1].size == 216
        assert loader.partitions[2].utt_ids == ['utt-4', 'utt-5']
        assert loader.partitions[2].utt_lengths == [(2,), (5,)]
        assert loader.partitions[2].size == 168

    def test_reload_creates_no_partition_with_no_utterances(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c1.open()

        loader = partitioning.PartitioningContainerLoader([], c1, '250', shuffle=False)

        assert len(loader.partitions) == 0

    def test_reload_creates_different_partitions_on_second_run(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c1.open()
        c1.set('utt-1', np.random.random((6, 6)).astype(np.float32))
        c1.set('utt-2', np.random.random((2, 6)).astype(np.float32))
        c1.set('utt-3', np.random.random((9, 6)).astype(np.float32))
        c1.set('utt-4', np.random.random((2, 6)).astype(np.float32))
        c1.set('utt-5', np.random.random((5, 6)).astype(np.float32))

        loader = partitioning.PartitioningContainerLoader(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                                          c1, '250', shuffle=True, seed=100)

        partitions_one = loader.partitions
        loader.reload()
        partitions_two = loader.partitions

        len_changed = len(partitions_one) == len(partitions_two)

        if len_changed:
            assert True
        else:
            utt_ids_changed = False

            for x, y in zip(partitions_one, partitions_two):
                if x.utt_ids != y.utt_ids:
                    utt_ids_changed = True

            assert utt_ids_changed

    def test_load_partition_data(self, tmpdir):
        c1 = containers.Container(os.path.join(tmpdir.strpath, 'c1.h5'))
        c1.open()
        utt_1_data = np.random.random((6, 6)).astype(np.float32)
        utt_2_data = np.random.random((2, 6)).astype(np.float32)
        utt_3_data = np.random.random((9, 6)).astype(np.float32)
        utt_4_data = np.random.random((2, 6)).astype(np.float32)
        utt_5_data = np.random.random((5, 6)).astype(np.float32)
        c1.set('utt-1', utt_1_data)
        c1.set('utt-2', utt_2_data)
        c1.set('utt-3', utt_3_data)
        c1.set('utt-4', utt_4_data)
        c1.set('utt-5', utt_5_data)

        loader = partitioning.PartitioningContainerLoader(['utt-1', 'utt-2', 'utt-3', 'utt-4', 'utt-5'],
                                                          c1, '250', shuffle=False)

        part_1 = loader.load_partition_data(0)
        assert part_1.info.utt_ids == ['utt-1', 'utt-2']
        assert np.allclose(part_1.utt_data[0], utt_1_data)
        assert np.allclose(part_1.utt_data[1], utt_2_data)

        part_2 = loader.load_partition_data(1)
        assert part_2.info.utt_ids == ['utt-3']
        assert np.allclose(part_2.utt_data[0], utt_3_data)

        part_3 = loader.load_partition_data(2)
        assert part_3.info.utt_ids == ['utt-4', 'utt-5']
        assert np.allclose(part_3.utt_data[0], utt_4_data)
        assert np.allclose(part_3.utt_data[1], utt_5_data)


class TestPartitionInfo:

    def test_total_length_for_single_container(self):
        info = partitioning.PartitionInfo()
        info.utt_lengths = [(1,), (9,), (13,)]

        assert info.total_lengths() == (23,)

    def test_total_length_for_multiple_containers(self):
        info = partitioning.PartitionInfo()
        info.utt_lengths = [(1, 4), (9, 5), (13, 8)]

        assert info.total_lengths() == (23, 17)


class TestPartitioningFeatureIterator(object):

    def test_next_emits_no_features_if_file_is_empty(self, tmpdir):
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')

        features = tuple(PartitioningFeatureIterator(file, 120))
        assert 0 == len(features)

    def test_next_emits_no_features_if_data_set_is_empty(self, tmpdir):
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=np.array([]))

        features = tuple(PartitioningFeatureIterator(file, 120))
        assert 0 == len(features)

    def test_next_emits_all_features_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=False))

        assert 5 == len(features)

        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[0])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[1])
        self.assert_features_equal(('utt-2', 0, [0.3, 0.3, 0.3, 0.3, 0.3]), features[2])
        self.assert_features_equal(('utt-2', 1, [0.4, 0.4, 0.4, 0.4, 0.4]), features[3])
        self.assert_features_equal(('utt-2', 2, [0.5, 0.5, 0.5, 0.5, 0.5]), features[4])

    def test_next_emits_all_features_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=True, seed=16))

        assert 6 == len(features)

        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[0])
        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[1])
        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[2])
        self.assert_features_equal(('utt-2', 2, [0.5, 0.5, 0.5, 0.5, 0.5]), features[3])
        self.assert_features_equal(('utt-2', 0, [0.3, 0.3, 0.3, 0.3, 0.3]), features[4])
        self.assert_features_equal(('utt-2', 1, [0.4, 0.4, 0.4, 0.4, 0.4]), features[5])

    def test_next_emits_features_only_from_included_ds_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=False, includes=['utt-1', 'utt-3', 'unknown']))

        assert 4 == len(features)

        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[0])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[1])
        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[2])
        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[3])

    def test_next_emits_features_only_from_included_ds_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=True, seed=16,
                                                     includes=['utt-1', 'utt-3', 'unknown']))

        assert 4 == len(features)

        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[0])
        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[1])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[2])
        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[3])

    def test_next_emits_features_without_excluded_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=False, excludes=['utt-2', 'unknown']))

        assert 4 == len(features)

        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[0])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[1])
        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[2])
        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[3])

    def test_next_emits_features_without_excluded_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=True, seed=16, excludes=['utt-2', 'unknown']))

        assert 4 == len(features)

        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[0])
        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[1])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[2])
        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[3])

    def test_next_emits_features_only_from_included_ds_ignoring_filter_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=False, includes=['utt-1', 'utt-3', 'unknown'],
                                                     excludes=['utt-1']))

        assert 4 == len(features)

        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[0])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[1])
        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[2])
        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[3])

    def test_next_emits_features_only_from_included_ds_ignoring_filter_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        features = tuple(PartitioningFeatureIterator(file, 120, shuffle=True, seed=16,
                                                     includes=['utt-1', 'utt-3', 'unknown'], excludes=['utt-1']))

        assert 4 == len(features)

        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[0])
        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[1])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[2])
        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[3])

    def test_next_emits_all_features_if_partition_spans_multiple_data_sets_in_sequential_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        iterator = PartitioningFeatureIterator(file, 240, shuffle=False)
        features = tuple(iterator)

        assert 7 == len(features)

        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[0])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[1])
        self.assert_features_equal(('utt-2', 0, [0.3, 0.3, 0.3, 0.3, 0.3]), features[2])
        self.assert_features_equal(('utt-2', 1, [0.4, 0.4, 0.4, 0.4, 0.4]), features[3])
        self.assert_features_equal(('utt-2', 2, [0.5, 0.5, 0.5, 0.5, 0.5]), features[4])
        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[5])
        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[6])

    def test_next_emits_all_features_if_partition_spans_multiple_data_sets_in_random_order(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        iterator = PartitioningFeatureIterator(file, 240, shuffle=True, seed=42)
        features = tuple(iterator)

        assert 7 == len(features)

        self.assert_features_equal(('utt-3', 1, [0.7, 0.7, 0.7, 0.7, 0.7]), features[0])
        self.assert_features_equal(('utt-1', 0, [0.1, 0.1, 0.1, 0.1, 0.1]), features[1])
        self.assert_features_equal(('utt-1', 1, [0.2, 0.2, 0.2, 0.2, 0.2]), features[2])
        self.assert_features_equal(('utt-3', 0, [0.6, 0.6, 0.6, 0.6, 0.6]), features[3])
        self.assert_features_equal(('utt-2', 0, [0.3, 0.3, 0.3, 0.3, 0.3]), features[4])
        self.assert_features_equal(('utt-2', 2, [0.5, 0.5, 0.5, 0.5, 0.5]), features[5])
        self.assert_features_equal(('utt-2', 1, [0.4, 0.4, 0.4, 0.4, 0.4]), features[6])

    def test_partitioning_empty_file_emits_zero_partitions(self, tmpdir):
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')

        iterator = PartitioningFeatureIterator(file, 100, shuffle=True, seed=42)

        assert 0 == len(iterator._partitions)

    def test_partitioning_empty_ds_emits_zero_partitions(self, tmpdir):
        ds1 = np.array([])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)

        iterator = PartitioningFeatureIterator(file, 100, shuffle=True, seed=42)

        assert 0 == len(iterator._partitions)

    def test_partitioning_all_data_fits_exactly_in_one(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        ds3 = np.array([[0.6, 0.6, 0.6, 0.6, 0.6], [0.7, 0.7, 0.7, 0.7, 0.7]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)
        file.create_dataset('utt-3', data=ds3)

        iterator = PartitioningFeatureIterator(file, 320, shuffle=True, seed=42)

        assert 1 == len(iterator._partitions)
        assert (('utt-1', 0), ('utt-3', 2)) in iterator._partitions

    def test_partitioning_partition_size_not_divisible_by_record_size_without_remainder(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)

        iterator = PartitioningFeatureIterator(file, 81, shuffle=True, seed=42)  # one byte more than ds1

        assert 3 == len(iterator._partitions)
        assert (('utt-1', 1), ('utt-1', 2)) in iterator._partitions
        assert (('utt-2', 0), ('utt-2', 2)) in iterator._partitions
        assert (('utt-2', 2), ('utt-1', 1)) in iterator._partitions

    def test_partitioning_remaining_space_filled_with_next_data_set(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)

        iterator = PartitioningFeatureIterator(file, 120, shuffle=True, seed=32)

        assert 2 == len(iterator._partitions)
        assert (('utt-1', 0), ('utt-2', 1)) in iterator._partitions
        assert (('utt-2', 1), ('utt-2', 3)) in iterator._partitions

    def test_partitioning_copes_with_varying_record_sizes(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1], [0.2, 0.2, 0.2, 0.2, 0.2]])
        ds2 = np.array([[0.3, 0.3, 0.3, 0.3, 0.3, 0.3], [0.4, 0.4, 0.4, 0.4, 0.4, 0.4], [0.5, 0.5, 0.5, 0.5, 0.5, 0.5]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)
        file.create_dataset('utt-2', data=ds2)

        iterator = PartitioningFeatureIterator(file, 120, shuffle=True, seed=12)

        assert 3 == len(iterator._partitions)
        assert (('utt-1', 0), ('utt-1', 2)) in iterator._partitions
        assert (('utt-2', 0), ('utt-2', 2)) in iterator._partitions
        assert (('utt-2', 2), ('utt-2', 3)) in iterator._partitions

    def test_partitioning_raises_error_if_record_bigger_than_partition_size(self, tmpdir):
        ds1 = np.array([[0.1, 0.1, 0.1, 0.1, 0.1]])
        file_path = os.path.join(tmpdir.strpath, 'features.h5')
        file = h5py.File(file_path, 'w')
        file.create_dataset('utt-1', data=ds1)

        with pytest.raises(ValueError):
            PartitioningFeatureIterator(file, 1)

    @staticmethod
    def assert_features_equal(expected, actual):
        if expected[0] != actual[0] or expected[1] != actual[1] or not np.allclose(expected[2], actual[2]):
            raise AssertionError('Expected {0} but got {1} instead'.format(expected, actual))



import pytest

from audiomate.utils import units


def test_partition_size_in_bytes_specified_as_int():
    assert 1024 == units.parse_storage_size(1024)


def test_partition_size_in_bytes():
    assert 1024 == units.parse_storage_size('1024')


def test_partition_size_in_kibibytes():
    assert 2 * 1024 == units.parse_storage_size('2k')


def test_partition_size_in_mebibytes():
    assert 2 * 1024 * 1024 == units.parse_storage_size('2m')


def test_partition_size_in_gibibytes():
    assert 2 * 1024 * 1024 * 1024 == units.parse_storage_size('2g')


def test_partition_size_in_gibibytes_with_capital_g():
    assert 2 * 1024 * 1024 * 1024 == units.parse_storage_size('2G')


def test_partition_size_fractions_of_bytes_are_ignored():
    assert 1 == units.parse_storage_size('1.1')


def test_partition_size_half_a_gibibyte():
    assert 512 * 1024 * 1024 == units.parse_storage_size('0.5g')


class TestUnits:
    def test_seconds_to_sample(self):
        assert units.seconds_to_sample(0, 16000) == 0
        assert units.seconds_to_sample(4.5, 22050) == 99225

    def test_sample_to_seconds(self):
        assert units.sample_to_seconds(0, 44100) == 0
        assert units.sample_to_seconds(3000, 8000) == 0.375


class TestFrameSettings:

    @pytest.mark.parametrize('frame_size,hop_size,num_samples,num_frames', [
        (4, 2, 12, 5),
        (4, 2, 13, 6),
        (4, 2, 15, 7),
        (4, 2, 16, 7),
        (4096, 2048, 41523, 20),
        (32000, 16000, 240000, 14)
    ])
    def test_num_frames(self, frame_size, hop_size, num_samples, num_frames):
        f = units.FrameSettings(frame_size, hop_size)
        assert f.num_frames(num_samples) == num_frames

    @pytest.mark.parametrize('frame_size,hop_size,frame_index,start_sample,end_sample', [
        (400, 160, 0, 0, 400),
        (400, 160, 5, 800, 1200),
    ])
    def test_frame_to_sample(self, frame_size, hop_size, frame_index, start_sample, end_sample):
        f = units.FrameSettings(frame_size, hop_size)
        assert f.frame_to_sample(frame_index) == (start_sample, end_sample)

    @pytest.mark.parametrize('frame_size,hop_size,sample_index,start_frame,end_frame', [
        (4, 2, 0, 0, 1),
        (4, 2, 1, 0, 1),
        (4, 2, 2, 0, 2),
        (4, 2, 5, 1, 3)
    ])
    def test_sample_to_frame_range(self, frame_size, hop_size, sample_index, start_frame, end_frame):
        f = units.FrameSettings(frame_size, hop_size)
        assert f.sample_to_frame_range(sample_index) == (start_frame, end_frame)

    @pytest.mark.parametrize('frame_size,hop_size,frame_index,sr,start,end', [
        (400, 160, 13, 16000, 0.13, 0.155)
    ])
    def test_frame_to_seconds(self, frame_size, hop_size, frame_index, sr, start, end):
        f = units.FrameSettings(frame_size, hop_size)
        assert f.frame_to_seconds(frame_index, sr) == (start, end)

    @pytest.mark.parametrize('frame_size,hop_size,start_time,end_time,sr,start_index,end_index', [
        (400, 160, 0.37, 2.84, 8000, 17, 142),
        (32000, 16000, 0.0, 5.0, 16000, 0, 5),
        (32000, 16000, 13.0, 15.0, 16000, 12, 15),
    ])
    def test_time_range_to_frame_range(self, frame_size, hop_size, start_time, end_time, sr, start_index, end_index):
        f = units.FrameSettings(frame_size, hop_size)
        assert f.time_range_to_frame_range(start_time, end_time, sr) == (start_index, end_index)



"""
This module contains functions for reading and writing json files.
"""

import json


def write_json_to_file(path, data):
    """
    Writes data as json to file.

    Parameters:
        path (str): Path to write to
        data (dict, list): Data
    """
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f)


def read_json_file(path):
    """
    Reads and return the data from the json file at the given path.

    Parameters:
        path (str): Path to read

    Returns:
        dict,list: The read json as dict/list.
    """

    with open(path, 'r', encoding='utf-8') as f:
        data = json.load(f)

    return data



import os
import glob

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus import subset
from audiomate.utils import textfile
from . import base


INVALID_UTTS = [
    'common_voice_de_17876483',
    'common_voice_de_17876495',
    'common_voice_de_17876535',
    'common_voice_de_17876537',
    'common_voice_de_17876539',
    'common_voice_de_17876540',
    'common_voice_de_17876541',
    'common_voice_de_17876542',
    'common_voice_de_17876543',
    'common_voice_de_17876544',
    'common_voice_de_17876545',
    'common_voice_de_17876546',
    'common_voice_de_17876547',
    'common_voice_de_17876548',
    'common_voice_de_17876549',
    'common_voice_de_17876550',
    'common_voice_de_17876551',
    'common_voice_de_17876552',
    'common_voice_de_17876553',
    'common_voice_de_17876554',
    'common_voice_de_17876555',
    'common_voice_de_17876556',
    'common_voice_de_17876557',
    'common_voice_de_17876304'
]


class CommonVoiceReader(base.CorpusReader):
    """
    Reader for the Common Voice Corpus.

    .. seealso::

       `Common-Voice <https://voice.mozilla.org/>`_
          Project Page
    """

    @classmethod
    def type(cls):
        return 'common-voice'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)
        subset_ids = CommonVoiceReader.get_subset_ids(path)

        for subset_idx in subset_ids:
            CommonVoiceReader.load_subset(corpus, path, subset_idx)

        return corpus

    @staticmethod
    def get_subset_ids(path):
        """ Return a list with ids of all available subsets (based on existing csv-files). """
        all = []

        for path in glob.glob(os.path.join(path, '*.tsv')):
            file_name = os.path.split(path)[1]
            basename = os.path.splitext(file_name)[0]

            # We don't want to include the invalidated files
            # since there maybe corrupt files
            if basename != 'invalidated':
                all.append(basename)

        return all

    @staticmethod
    def load_subset(corpus, path, subset_idx):
        """ Load subset into corpus. """
        csv_file = os.path.join(path, '{}.tsv'.format(subset_idx))
        subset_utt_ids = []

        entries = textfile.read_separated_lines_generator(
            csv_file,
            separator='\t',
            max_columns=8,
            ignore_lines_starting_with=['client_id'],
            keep_empty=True
        )

        for entry in entries:
            file_idx = CommonVoiceReader.create_assets_if_needed(
                corpus,
                path,
                entry
            )

            if file_idx is not None:
                subset_utt_ids.append(file_idx)

        filter = subset.MatchingUtteranceIdxFilter(utterance_idxs=set(subset_utt_ids))
        subview = subset.Subview(corpus, filter_criteria=[filter])
        corpus.import_subview(subset_idx, subview)

    @staticmethod
    def create_assets_if_needed(corpus, path, entry):
        """ Create File/Utterance/Issuer, if they not already exist and return utt-idx. """
        file_name = entry[1]
        file_idx, _ = os.path.splitext(file_name)

        if file_idx in INVALID_UTTS:
            return None

        if file_idx not in corpus.utterances.keys():
            speaker_idx = entry[0]
            transcription = entry[2]

            if len(entry) >= 6:
                age = CommonVoiceReader.map_age(entry[5])
            else:
                age = issuers.AgeGroup.UNKNOWN

            if len(entry) >= 7:
                gender = CommonVoiceReader.map_gender(entry[6])
            else:
                gender = issuers.Gender.UNKNOWN

            file_path = os.path.join(path, 'clips', file_name)
            corpus.new_file(file_path, file_idx)

            if speaker_idx in corpus.issuers.keys():
                issuer = corpus.issuers[speaker_idx]
            else:
                issuer = issuers.Speaker(speaker_idx, gender=gender, age_group=age)
                corpus.import_issuers(issuer)

            utterance = corpus.new_utterance(file_idx, file_idx, issuer.idx)
            utterance.set_label_list(
                annotations.LabelList.create_single(
                    transcription,
                    idx=audiomate.corpus.LL_WORD_TRANSCRIPT
                )
            )

        return file_idx

    @staticmethod
    def map_age(age):
        """ Map age to correct age-group. """

        if age in [None, '']:
            return issuers.AgeGroup.UNKNOWN
        elif age == 'teens':
            return issuers.AgeGroup.YOUTH
        elif age in ['sixties', 'seventies', 'eighties', 'nineties']:
            return issuers.AgeGroup.SENIOR
        else:
            return issuers.AgeGroup.ADULT

    @staticmethod
    def map_gender(gender):
        """ Map gender to correct value. """

        if gender == 'male':
            return issuers.Gender.MALE
        elif gender == 'female':
            return issuers.Gender.FEMALE
        else:
            return issuers.Gender.UNKNOWN



import pytest

from audiomate.utils import misc


@pytest.mark.parametrize('first_start,first_end,second_start,second_end,overlap', [
    (1.5, 4.2, 3.5, 7.0, 0.7),
    (1.5, 4.2, 0.4, 2.0, 0.5),
    (1.5, 4.2, 1.7, 2.24, 0.54),
    (1.5, 4.2, 0.3, 7.3, 2.7),
    (1.5, 1.7, 1.7, 2.24, 0.0),
    (2.8, 2.9, 1.7, 2.24, 0.0)
])
def test_length_of_overlap(first_start, first_end, second_start, second_end, overlap):
    actual_overlap = misc.length_of_overlap(first_start, first_end, second_start, second_end)
    assert actual_overlap == pytest.approx(overlap)



"""
This module contains any functions for working with text/strings/punctuation.
"""

import re


def remove_punctuation(text, exceptions=[]):
    """
    Return a string with punctuation removed.

    Parameters:
        text (str): The text to remove punctuation from.
        exceptions (list): List of symbols to keep in the given text.

    Return:
        str: The input text without the punctuation.
    """

    all_but = [
        r'\w',
        r'\s'
    ]

    all_but.extend(exceptions)

    pattern = '[^{}]'.format(''.join(all_but))

    return re.sub(pattern, '', text)


def starts_with_prefix_in_list(text, prefixes):
    """
    Return True if the given string starts with one of the prefixes in the given list, otherwise
    return False.

    Arguments:
        text (str): Text to check for prefixes.
        prefixes (list): List of prefixes to check for.

    Returns:
        bool: True if the given text starts with any of the given prefixes, otherwise False.
    """
    for prefix in prefixes:
        if text.startswith(prefix):
            return True
    return False



import pytest

from audiomate.annotations import Label, LabelList


class TestLabelList:

    def test_equal(self):
        ll_a = LabelList(idx='test', labels=[
            Label('c', 12.05, 14.0),
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('b', 7.9, 9.3),
            Label('c', 9.0, 12.0),
        ])

        ll_b = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 7.9, 9.3),
            Label('c', 9.0, 12.0),
            Label('b', 4.0, 8.0),
            Label('c', 12.05, 14.0),
        ])

        assert ll_a == ll_b

    def test_is(self):
        ll_a = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('b', 7.9, 9.3),
            Label('c', 9.0, 12.0),
            Label('c', 12.05, 14.0),
        ])

        ll_b = ll_a

        assert ll_a is ll_b

    def test_iter(self):
        ll = LabelList(labels=[
            Label('some text'),
            Label('more text'),
            Label('text again'),
        ])

        labels = sorted(ll)

        assert labels[0].value == 'more text'
        assert labels[1].value == 'some text'
        assert labels[2].value == 'text again'

    def test_len(self):
        ll = LabelList(labels=[
            Label('some text'),
            Label('more text'),
            Label('text again'),
        ])

        assert len(ll) == 3

    def test_labels(self):
        ll = LabelList(labels=[
            Label('a'),
            Label('c'),
            Label('b'),
        ])

        labels = ll.labels
        assert isinstance(labels, list)
        assert sorted(labels) == [
            Label('a'),
            Label('b'),
            Label('c'),
        ]

    def test_start(self):
        ll = LabelList(idx='test', labels=[
            Label('c', 9.0, 12.0),
            Label('a', 0.2, 4.0),
            Label('b', 4.0, 8.0),
            Label('b', 7.9, 9.3),
            Label('c', 12.05, 14.0),
        ])

        assert ll.start == 0.2

    def test_end(self):
        ll = LabelList(idx='test', labels=[
            Label('c', 9.0, 12.0),
            Label('a', 0.2, 4.0),
            Label('b', 4.0, 8.0),
            Label('c', 12.05, 14.0),
            Label('b', 7.9, 9.3),
        ])

        assert ll.end == 14.0

    def test_end_returns_inf(self):
        ll = LabelList(idx='test', labels=[
            Label('c', 9.0, 12.0),
            Label('a', 0.2, 4.0),
            Label('b', 4.0, float('inf')),
            Label('c', 12.05, 14.0),
            Label('b', 7.9, 9.3),
        ])

        assert ll.end == float('inf')

    def test_total_length(self):
        ll = LabelList(labels=[
            Label('abc', 3.2, 4.5),
            Label('bg', 5.1, 8.9),
            Label('caaf', 7.2, 10.5),
            Label('yxva', 10.5, 14),
            Label('cy', 13, 14)
        ])

        assert ll.total_length == 15

    def test_add(self):
        ll = LabelList()

        assert len(ll) == 0

        label = Label('some text')
        ll.add(label)

        assert len(ll) == 1
        assert label.label_list == ll
        assert sorted(ll)[0] == label

    def test_addl(self):
        ll = LabelList()

        assert len(ll) == 0

        ll.addl('a', 12.3, 19.3)

        assert ll == LabelList(labels=[
            Label('a', 12.3, 19.3),
        ])

    def test_update(self):
        ll = LabelList()

        assert len(ll) == 0

        label_a = Label('some text')
        label_b = Label('more text')
        label_c = Label('text again')
        ll.update([label_a, label_b, label_c])

        assert len(ll) == 3
        assert label_a.label_list == ll
        assert label_b.label_list == ll
        assert label_c.label_list == ll

    def test_apply(self):
        ll = LabelList(labels=[
            Label('some text'),
            Label('more text'),
            Label('text again'),
        ])

        def apply_func(label):
            label.value = 'app {}'.format(label.value)

        ll.apply(apply_func)

        labels = sorted(ll)
        assert labels[0].value == 'app more text'
        assert labels[1].value == 'app some text'
        assert labels[2].value == 'app text again'

    def test_merge_overlaps(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('b', 7.9, 9.3),
            Label('c', 9.0, 12.0),
            Label('c', 12.05, 14.0),
        ])
        ll.merge_overlaps()

        expected = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 9.3),
            Label('c', 9.0, 12.0),
            Label('c', 12.05, 14.0),
        ])

        assert ll == expected

    def test_merge_overlaps_with_threshold(self):
        ll = LabelList(idx='test', labels=[
            Label('c', 9.0, 12.0),
            Label('c', 12.05, 14.0),
        ])

        ll.merge_overlaps(threshold=0.1)

        expected = LabelList(idx='test', labels=[
            Label('c', 9.0, 14.0),
        ])

        assert ll == expected

    def test_merge_overlaps_with_multiple_consecutive_overlapping(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('b', 7.9, 9.3),
            Label('b', 8.9, 10.9),
            Label('c', 9.0, 12.0)
        ])

        ll.merge_overlaps()

        expected = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 10.9),
            Label('c', 9.0, 12.0),
        ])

        assert ll == expected

    def test_merge_overlaps_with_multiple_overlapping(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('b', 6.2, 9.3),
            Label('b', 7.1, 10.9),
            Label('c', 9.0, 12.0),
        ])

        ll.merge_overlaps()

        expected = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 10.9),
            Label('c', 9.0, 12.0),
        ])

        assert ll == expected

    def test_merge_overlaps_when_fully_contained(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('b', 5.0, 7.3),
            Label('b', 7.9, 8.7),
            Label('c', 9.0, 12.0),
        ])

        ll.merge_overlaps()

        expected = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.7),
            Label('c', 9.0, 12.0),
        ])

        assert ll == expected

    def test_label_total_duration(self):
        ll = LabelList(labels=[
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14),
            Label('c', 13, 14)
        ])

        res = ll.label_total_duration()

        assert res['a'] == pytest.approx(4.8)
        assert res['b'] == pytest.approx(3.8)
        assert res['c'] == pytest.approx(4.3)

    def test_label_values(self):
        ll = LabelList(labels=[
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14),
            Label('c', 13, 14)
        ])

        assert ll.label_values() == ['a', 'b', 'c']

    def test_label_count(self):
        ll = LabelList(labels=[
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14),
            Label('c', 13, 14)
        ])

        res = ll.label_count()

        assert 2 == res['a']
        assert 1 == res['b']
        assert 2 == res['c']

    def test_all_tokens(self):
        ll = LabelList(labels=[
            Label('some text'),
            Label('more text'),
            Label('text again'),
        ])

        assert sorted(ll.all_tokens()) == [
            'again',
            'more',
            'some',
            'text',
        ]

    def test_join(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('c', 9.0, 12.0)
        ])

        assert ll.join() == 'a b c'

    def test_join_with_custom_delimiter(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('c', 9.0, 12.0)
        ])

        assert ll.join(delimiter=' - ') == 'a - b - c'

    def test_join_raises_error_if_overlap_is_higher_than_threshold(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 3.8, 8.0),
            Label('c', 9.0, 12.0)
        ])

        with pytest.raises(ValueError):
            ll.join(overlap_threshold=0.1)

    def test_join_raises_error_if_overlap_is_higher_than_threshold_given_an_endless_label(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.5, float('inf')),
            Label('c', 9.0, 12.0)
        ])

        with pytest.raises(ValueError):
            ll.join(overlap_threshold=0.1)

    def test_tokenized(self):
        ll = LabelList(idx='test', labels=[
            Label('a u t', 0.0, 4.0),
            Label('b x', 4.0, 8.0),
            Label('c', 9.0, 12.0)
        ])

        assert ll.tokenized() == ['a', 'u', 't', 'b', 'x', 'c']

    def test_tokenized_raises_error_if_overlap_is_higher_than_threshold(self):
        ll = LabelList(idx='test', labels=[
            Label('a u t', 0.0, 4.0),
            Label('b x', 3.85, 8.0),
            Label('c', 9.0, 12.0)
        ])

        with pytest.raises(ValueError):
            ll.tokenized()

    def test_tokenized_raises_error_if_overlap_is_higher_than_threshold_given_an_endless_label(self):
        ll = LabelList(idx='test', labels=[
            Label('a u t', 0.0, 4.0),
            Label('b x', 4.5, float('inf')),
            Label('c', 9.0, 12.0)
        ])

        with pytest.raises(ValueError):
            ll.tokenized()

    def test_separated(self):
        ll = LabelList(idx='a', labels=[
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14),
            Label('c', 13, 14)
        ])

        res = ll.separated()

        assert res['a'] == LabelList(idx='a', labels=[
            Label('a', 3.2, 4.5),
            Label('a', 10.5, 14)
        ])
        assert res['b'] == LabelList(idx='a', labels=[
            Label('b', 5.1, 8.9)
        ])
        assert res['c'] == LabelList(idx='a', labels=[
            Label('c', 7.2, 10.5),
            Label('c', 13, 14)
        ])

    def test_labels_in_range(self):
        ll = LabelList(labels=[
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14),
            Label('c', 13, 14)
        ])

        in_range = ll.labels_in_range(8.2, 12.5)

        assert sorted(in_range) == [
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14)
        ]

    def test_labels_in_range_returns_only_fully_included(self):
        ll = LabelList(labels=[
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14),
            Label('c', 13, 15)
        ])

        in_range = ll.labels_in_range(7.2, 14.99, fully_included=True)

        assert sorted(in_range) == [
            Label('c', 7.2, 10.5),
            Label('a', 10.5, 14),
        ]

    def test_ranges(self):
        labels = [
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('d', 10.5, 14)
        ]
        ll = LabelList(labels=labels)

        ranges = ll.ranges()

        r = next(ranges)
        assert 3.2 == r[0]
        assert 4.5 == r[1]
        assert labels[0] in r[2]

        r = next(ranges)
        assert 5.1 == r[0]
        assert 7.2 == r[1]
        assert labels[1] in r[2]

        r = next(ranges)
        assert 7.2 == r[0]
        assert 8.9 == r[1]
        assert labels[1] in r[2]
        assert labels[2] in r[2]

        r = next(ranges)
        assert 8.9 == r[0]
        assert 10.5 == r[1]
        assert labels[2] in r[2]

        r = next(ranges)
        assert 10.5 == r[0]
        assert 14 == r[1]
        assert labels[3] in r[2]

        with pytest.raises(StopIteration):
            next(ranges)

    def test_ranges_with_empty(self):
        labels = [
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9),
            Label('c', 7.2, 10.5),
            Label('d', 10.5, 14)
        ]
        ll = LabelList(labels=labels)

        ranges = ll.ranges(yield_ranges_without_labels=True)

        r = next(ranges)
        assert 3.2 == r[0]
        assert 4.5 == r[1]
        assert labels[0] in r[2]

        r = next(ranges)
        assert 4.5 == r[0]
        assert 5.1 == r[1]
        assert 0 == len(r[2])

        r = next(ranges)
        assert 5.1 == r[0]
        assert 7.2 == r[1]
        assert labels[1] in r[2]

        r = next(ranges)
        assert 7.2 == r[0]
        assert 8.9 == r[1]
        assert labels[1] in r[2]
        assert labels[2] in r[2]

        r = next(ranges)
        assert 8.9 == r[0]
        assert 10.5 == r[1]
        assert labels[2] in r[2]

        r = next(ranges)
        assert 10.5 == r[0]
        assert 14 == r[1]
        assert labels[3] in r[2]

        with pytest.raises(StopIteration):
            next(ranges)

    def test_ranges_include_labels(self):
        labels = [
            Label('a', 3.2, 4.5),
            Label('b', 5.1, 8.9)
        ]
        ll = LabelList(labels=labels)

        ranges = ll.ranges(include_labels=['a'])

        r = next(ranges)
        assert 3.2 == r[0]
        assert 4.5 == r[1]
        assert labels[0] in r[2]

        with pytest.raises(StopIteration):
            next(ranges)

    def test_ranges_zero_to_end(self):
        labels = [
            Label('a', 0, float('inf')),
            Label('b', 5.1, 8.9)
        ]
        ll = LabelList(labels=labels)

        ranges = ll.ranges()

        r = next(ranges)
        assert 0 == r[0]
        assert 5.1 == r[1]
        assert labels[0] in r[2]

        r = next(ranges)
        assert 5.1 == r[0]
        assert 8.9 == r[1]
        assert labels[0] in r[2]
        assert labels[1] in r[2]

        r = next(ranges)
        assert 8.9 == r[0]
        assert float('inf') == r[1]
        assert labels[0] in r[2]

        with pytest.raises(StopIteration):
            next(ranges)

    def test_ranges_with_same_start_times(self):
        labels = [
            Label('a', 1.2, 1.3),
            Label('b', 1.2, 5.6)
        ]
        ll = LabelList(labels=labels)

        ranges = ll.ranges()

        r = next(ranges)
        assert r[0] == 1.2
        assert r[1] == 1.3
        assert len(r[2]) == 2
        assert labels[0] in r[2]
        assert labels[1] in r[2]

        r = next(ranges)
        assert r[0] == 1.3
        assert r[1] == 5.6
        assert len(r[2]) == 1
        assert labels[1] in r[2]

        with pytest.raises(StopIteration):
            next(ranges)

    def test_split(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('c', 9.0, 12.0)
        ])

        res = ll.split([1.9, 6.2, 10.5])

        assert res == [
            LabelList(idx='test', labels=[
                Label('a', 0.0, 1.9)
            ]),
            LabelList(idx='test', labels=[
                Label('a', 1.9, 4.0),
                Label('b', 4.0, 6.2),
            ]),
            LabelList(idx='test', labels=[
                Label('b', 6.2, 8.0),
                Label('c', 9.0, 10.5),
            ]),
            LabelList(idx='test', labels=[
                Label('c', 10.5, 12.0),
            ]),
        ]

    def test_split_unsorted_label_list(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('c', 9.0, 12.0),
            Label('b', 4.0, 8.0)
        ])

        res = ll.split([1.9, 6.2, 10.5])

        assert res == [
            LabelList(idx='test', labels=[
                Label('a', 0.0, 1.9),
            ]),
            LabelList(idx='test', labels=[
                Label('a', 1.9, 4.0),
                Label('b', 4.0, 6.2),
            ]),
            LabelList(idx='test', labels=[
                Label('b', 6.2, 8.0),
                Label('c', 9.0, 10.5),
            ]),
            LabelList(idx='test', labels=[
                Label('c', 10.5, 12.0),
            ]),
        ]

    def test_split_label_within_cutting_points_is_included(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('c', 9.0, 12.0)
        ])

        res = ll.split([1.9, 10.5])

        assert len(res[1]) == 3
        assert sorted(res[1])[1].value == 'b'
        assert sorted(res[1])[1].start == 4.0
        assert sorted(res[1])[1].end == 8.0

    def test_split_with_endless_label(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('c', 4.0, float('inf'))
        ])

        res = ll.split([1.9, 10.5])

        assert len(res) == 3

        assert len(res[2]) == 1
        assert sorted(res[1])[1].value == 'c'
        assert sorted(res[1])[1].start == 4.0
        assert sorted(res[1])[1].end == 10.5
        assert sorted(res[2])[0].value == 'c'
        assert sorted(res[2])[0].start == 10.5
        assert sorted(res[2])[0].end == float('inf')

    def test_split_with_cutting_point_after_last_label(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('c', 4.0, 8.9)
        ])

        res = ll.split([10.5])

        assert len(res) == 2
        assert len(res[0]) == 2
        assert len(res[1]) == 0

    def test_split_cutting_point_on_boundary_doesnot_split_label(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 9.0),
            Label('c', 9.0, 12.0)
        ])

        res = ll.split([9.0])

        assert len(res) == 2

        assert len(res[0]) == 1
        assert len(res[1]) == 1

    def test_split_without_cutting_points_raises_error(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 9.0),
            Label('c', 9.0, 12.0)
        ])

        with pytest.raises(ValueError):
            ll.split([])

    def test_split_with_shifting_start_and_endtime(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 9.0),
            Label('c', 9.0, 12.0)
        ])

        res = ll.split([4.2], shift_times=True)

        assert len(res) == 2

        assert len(res[0]) == 1
        assert sorted(res[0])[0].value == 'a'
        assert sorted(res[0])[0].start == 0.0
        assert sorted(res[0])[0].end == 4.2

        assert len(res[1]) == 2
        assert sorted(res[1])[0] == Label('a', 0.0, 4.8)
        assert sorted(res[1])[1] == Label('c', 4.8, 7.8)

    def test_split_first_label_not_splitted(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 9.0),
            Label('c', 9.0, 12.0)
        ])

        res = ll.split([11.2], shift_times=True)

        assert len(res) == 2

        assert len(res[0]) == 2
        assert sorted(res[0])[0] == Label('a', 0.0, 9.0)
        assert sorted(res[0])[1] == Label('c', 9.0, 11.2)

        assert len(res[1]) == 1
        assert sorted(res[1])[0] == Label('c', 0.0, pytest.approx(0.8))

    def test_split_single_label_that_doesnt_start_at_zero(self):
        ll = LabelList(idx='test', labels=[
            Label('c', 8.0, 12.0)
        ])

        res = ll.split([11.2], shift_times=True)

        assert len(res) == 2

        assert len(res[0]) == 1
        assert sorted(res[0])[0] == Label('c', 8.0, 11.2)

        assert len(res[1]) == 1
        assert sorted(res[1])[0] == Label('c', 0.0, pytest.approx(0.8))

    def test_split_with_overlap(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 0.0, 4.0),
            Label('b', 4.0, 8.0),
            Label('c', 9.0, 12.0)
        ])

        res = ll.split([1.9, 6.2, 10.5], overlap=2.0)

        assert res == [
            LabelList(idx='test', labels=[
                Label('a', 0.0, 3.9)
            ]),
            LabelList(idx='test', labels=[
                Label('a', 0, 4.0),
                Label('b', 4.0, 8.0)
            ]),
            LabelList(idx='test', labels=[
                Label('b', 4.2, 8.0),
                Label('c', 9.0, 12.0)]
            ),
            LabelList(idx='test', labels=[
                Label('c', 9.0, 12.0)
            ]),
        ]

    def test_split_with_overlap_label_only_touches_overlap_time(self):
        ll = LabelList(idx='test', labels=[
            Label('a', 2.0, 5.0),
            Label('b', 7.0, 8.0),
        ])

        res = ll.split([6.2, 10.5], overlap=2.0)

        assert res == [
            LabelList(idx='test', labels=[
                Label('a', 2.0, 5.0),
                Label('b', 7.0, 8.0),
            ]),

            LabelList(idx='test', labels=[
                Label('a', 4.2, 5.0),
                Label('b', 7.0, 8.0),
            ]),
            LabelList(idx='test', labels=[])
        ]

    def test_split_with_shift_and_overlap(self):
        ll = LabelList('test', labels=[
            Label('alpha', start=0.0, end=30.0),
            Label('bravo', start=20.0, end=42.0)
        ])

        res = ll.split([12.0, 24.0], shift_times=True, overlap=2.0)

        for x in res:
            print(x.labels)

        assert res == [
            LabelList(idx='test', labels=[
                Label('alpha', 0.0, 14.0),
            ]),

            LabelList(idx='test', labels=[
                Label('alpha', 0.0, 16.0),
                Label('bravo', 10.0, 16.0),
            ]),
            LabelList(idx='test', labels=[
                Label('alpha', 0.0, 8.0),
                Label('bravo', 0.0, 20.0),
            ])
        ]

    def test_create_single(self):
        ll = LabelList.create_single('bob')

        assert ll == LabelList(idx='default', labels=[
            Label('bob')
        ])

    def test_create_single_with_custom_idx(self):
        ll = LabelList.create_single('bob', idx='name')

        assert ll == LabelList(idx='name', labels=[
            Label('bob')
        ])

    def test_with_label_values(self):
        ll = LabelList.with_label_values([
            'a',
            'b',
            'c',
        ])

        assert ll == LabelList(labels=[
            Label('a'),
            Label('b'),
            Label('c'),
        ])

    def test_with_label_values_sets_correct_idx(self):
        ll = LabelList.with_label_values([
            'a',
            'b',
            'c',
        ], idx='letters')

        assert ll == LabelList(idx='letters', labels=[
            Label('a'),
            Label('b'),
            Label('c'),
        ])



from audiomate.corpus import io

from tests import resources
from . import reader_test as rt


class TestFolderReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('folder')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 7
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('empty', 'empty.wav'),
        rt.ExpFileTrack('wav_1', 'wav_1.wav'),
        rt.ExpFileTrack('wav_2', 'wav_2.wav'),
        rt.ExpFileTrack('wav_3', 'wav_3.wav'),
        rt.ExpFileTrack('wav_4', 'wav_4.wav'),
        rt.ExpFileTrack('wav_200_samples', 'wav_200_samples.wav'),
        rt.ExpFileTrack('wav_invalid', 'wav_invalid.wav'),
    ]

    EXPECTED_NUMBER_OF_UTTERANCES = 7
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('empty', 'empty', None, 0, float('inf')),
        rt.ExpUtterance('wav_1', 'wav_1', None, 0, float('inf')),
        rt.ExpUtterance('wav_2', 'wav_2', None, 0, float('inf')),
        rt.ExpUtterance('wav_3', 'wav_3', None, 0, float('inf')),
        rt.ExpUtterance('wav_4', 'wav_4', None, 0, float('inf')),
        rt.ExpUtterance('wav_200_samples', 'wav_200_samples', None, 0, float('inf')),
        rt.ExpUtterance('wav_invalid', 'wav_invalid', None, 0, float('inf')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 0
    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        return io.FolderReader().load(self.SAMPLE_PATH)



import os
import glob

import audiomate
from audiomate import annotations
from audiomate import issuers
from . import base
from . import downloader

MASTER_DOWNLOAD_URL = 'https://github.com/Jakobovski/free-spoken-digit-dataset/archive/master.zip'


class FreeSpokenDigitDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the Free-Spoken-Digit dataset.

    Args:
        url (str): The url to download the dataset from. If not given the default URL is used.
                   It is expected to be a zip file.
    """

    def __init__(self, url=None):
        if url is None:
            url = MASTER_DOWNLOAD_URL

        super(FreeSpokenDigitDownloader, self).__init__(
            url,
            move_files_up=True
        )

    @classmethod
    def type(cls):
        return 'free-spoken-digits'


class FreeSpokenDigitReader(base.CorpusReader):
    """
    Reader for the Free-Spoken-Digit Corpus.

    .. seealso::

       `Free-Spoken-Digit-Dataset <https://github.com/Jakobovski/free-spoken-digit-dataset>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'free-spoken-digits'

    def _check_for_missing_files(self, path):
        recordings_folder = os.path.join(path, 'recordings')

        if os.path.isdir(recordings_folder):
            return []
        else:
            return [recordings_folder]

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)

        for file_path in glob.glob(os.path.join(path, 'recordings', '*.wav')):
            file_idx = os.path.splitext(os.path.basename(file_path))[0]

            corpus.new_file(file_path, file_idx)

            idx_parts = file_idx.split('_')
            digit = idx_parts[0]
            issuer_idx = '_'.join(idx_parts[1:-1])

            if issuer_idx not in corpus.issuers.keys():
                issuer = issuers.Speaker(issuer_idx)
                corpus.import_issuers(issuer)

            utterance = corpus.new_utterance(file_idx, file_idx, issuer_idx)
            utterance.set_label_list(annotations.LabelList.create_single(str(digit),
                                                                         idx=audiomate.corpus.LL_WORD_TRANSCRIPT))

        return corpus



import contextlib
import h5py


class Container(object):
    """
    A container is a wrapper around a HDF5 file.
    In a container is used to store array-like data.
    Every array is associated with some idx/key.
    Every array (a dataset in h5py-terms) may have additional attributes.

    Args:
        path (str): Path where the HDF5 file is stored.
                    If the file doesn't exist, one is created.
        mode (str): Either 'r' for read-only, 'w' for truncate and write or
                    'a' for append. (default: 'a').

    Example:
        >>> ct = Container('/path/to/hdf5file')
        >>> with ct:
        >>>     ct.set('utt-1', np.array([1,2,3,4]))
        >>>     data = ct.get('utt-1')
        array([1, 2, 3, 4])
    """

    def __init__(self, path, mode='a'):
        if mode not in ['r', 'w', 'a']:
            raise ValueError('Invalid mode! Modes: [\'a\', \'r\', \'w\']')

        self.path = path
        self.mode = mode

        self._file = None

    def open(self, mode=None):
        """
        Open the container file.

        Args:
            mode (str): Either 'r' for read-only, 'w' for truncate and write or
                        'a' for append. (default: 'a').
                        If ``None``, uses ``self.mode``.
        """

        if mode is None:
            mode = self.mode
        elif mode not in ['r', 'w', 'a']:
            raise ValueError('Invalid mode! Modes: [\'a\', \'r\', \'w\']')

        if self._file is None:
            self._file = h5py.File(self.path, mode=mode)

    def close(self):
        """
        Close the container file if its open.
        """
        if self._file is not None:
            self._file.close()
            self._file = None

    def is_open(self, mode=None):
        """
        Return ``True``, if container is already open. ``False`` otherwise.
        """
        return self._file is not None

    @contextlib.contextmanager
    def open_if_needed(self, mode=None):
        """
        Convenience context-manager for the use with ``with``.
        Opens the container if not already done.
        Only closes the container if it was opened within this context.

        Args:
            mode (str): Either 'r' for read-only, 'w' for truncate and write or
                        'a' for append. (default: 'a').
                        If ``None``, uses ``self.mode``.
        """
        was_open = self.is_open()

        if not was_open:
            self.open(mode=mode)

        try:
            yield self
        finally:
            if not was_open:
                self.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def keys(self):
        """
        Return a list of keys for which an array is stored in the container.

        Returns:
            list: List of identifiers available in the container.

        Note:
            The container has to be opened in advance.
        """
        self.raise_error_if_not_open()

        return sorted(list(self._file.keys()))

    def get(self, key, mem_map=True):
        """
        Read and return the data stored for the given key.

        Args:
            key (str): The key to read the data from.
            mem_map (bool): If ``True`` returns the data as
                            memory-mapped array, otherwise a copy is returned.

        Note:
            The container has to be opened in advance.

        Returns:
            numpy.ndarray: The stored data.
        """
        self.raise_error_if_not_open()

        if key in self._file:
            data = self._file[key]

            if not mem_map:
                data = data[()]

            return data
        else:
            return None

    def set(self, key, data):
        """
        Set the given data to the container with the given key.
        Any existing data for the given key is discarded/overwritten.

        Args:
            key (str): A key to store the data for.
            data (numpy.ndarray): Array-like data.

        Note:
            The container has to be opened in advance.
        """
        self.raise_error_if_not_open()

        if key in self._file:
            del self._file[key]

        self._file.create_dataset(key, data=data)

    def append(self, key, data):
        """
        Append the given data to the data that already exists
        in the container for the given key.
        Only data with equal dimensions (except the first) are allowed,
        since they are concatenated/stacked along the first dimension.

        Args:
            key (str): Key to store data for.
            data (numpy.ndarray): Array-like data.
                                  Has to have the same dimension as
                                  the existing data after the first dimension.

        Note:
            The container has to be opened in advance.
            For appending to existing data the HDF5-Dataset has to be chunked,
            so it is not allowed to first add data via ``set``.
        """
        existing = self.get(key, mem_map=True)

        if existing is not None:
            num_existing = existing.shape[0]

            if existing.shape[1:] != data.shape[1:]:
                error_msg = (
                    'The data to append needs to'
                    'have the same dimensions ({}).'
                )
                raise ValueError(error_msg.format(existing.shape[1:]))

            existing.resize(num_existing + data.shape[0], 0)
            existing[num_existing:] = data
        else:
            max_shape = list(data.shape)
            max_shape[0] = None

            self._file.create_dataset(key, data=data,
                                      chunks=True, maxshape=max_shape)

    def remove(self, key):
        """
        Remove the data stored for the given key.

        Args:
            key (str): Key of the data to remove.

        Note:
            The container has to be opened in advance.
        """
        self.raise_error_if_not_open()

        if key in self._file:
            del self._file[key]

    def raise_error_if_not_open(self):
        """ Check if container is opened, raise error if not. """
        if self._file is None:
            raise ValueError('The container is not opened!')



import os

import pytest
import requests_mock

from audiomate.utils import download

from tests import resources


@pytest.fixture()
def sample_zip_data():
    with open(resources.get_resource_path(['sample_files', 'zip_sample.zip']), 'rb') as f:
        return f.read()


@pytest.fixture()
def sample_tar_bz2_path():
    return resources.get_resource_path(['sample_files', 'sentences.tar.bz2'])


@pytest.fixture()
def sample_zip_path():
    return resources.get_resource_path(['sample_files', 'zip_sample.zip'])


def test_download_file(sample_zip_data, tmpdir):
    dl_path = 'http://some.url/thezipfile.zip'
    target_path = os.path.join(tmpdir.strpath, 'target.zip')

    with requests_mock.Mocker() as mock:
        mock.get(dl_path, content=sample_zip_data)

        download.download_file(dl_path, target_path)

    assert os.path.isfile(target_path)

    with open(target_path, 'rb') as f:
        assert f.read() == sample_zip_data


def test_extract_zip(sample_zip_path, tmpdir):
    target_folder = tmpdir.strpath

    download.extract_zip(sample_zip_path, target_folder)

    assert os.path.isfile(os.path.join(target_folder, 'a.txt'))
    assert os.path.isfile(os.path.join(target_folder, 'data', 'dibsdadu.txt'))
    assert os.path.isfile(os.path.join(target_folder, 'data', 'babadu.txt'))


def test_extract_tar_bz2(sample_tar_bz2_path, tmpdir):
    download.extract_tar(sample_tar_bz2_path, tmpdir.strpath)

    assert os.path.isfile(os.path.join(tmpdir.strpath, 'sentences.csv'))



from audiomate import annotations
from audiomate import tracks
from audiomate.corpus.subset import subview

import pytest

from tests import resources


class TestMatchingUtteranceIdxFilter:

    def test_match(self):
        filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'a', 'b', 'd'})

        assert filter.match(tracks.Utterance('a', 'x'), None)
        assert filter.match(tracks.Utterance('b', 'x'), None)
        assert filter.match(tracks.Utterance('d', 'x'), None)
        assert not filter.match(tracks.Utterance('c', 'x'), None)
        assert not filter.match(tracks.Utterance('e', 'x'), None)

    def test_match_inverse(self):
        filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'a', 'b', 'd'}, inverse=True)

        assert not filter.match(tracks.Utterance('a', 'x'), None)
        assert not filter.match(tracks.Utterance('b', 'x'), None)
        assert not filter.match(tracks.Utterance('d', 'x'), None)
        assert filter.match(tracks.Utterance('c', 'x'), None)
        assert filter.match(tracks.Utterance('e', 'x'), None)

    def test_serialize(self):
        f = subview.MatchingUtteranceIdxFilter(utterance_idxs={'a', 'b', 'd'})
        assert f.serialize() == 'include,a,b,d'

    def test_serialize_inverse(self):
        f = subview.MatchingUtteranceIdxFilter(utterance_idxs={'a', 'b', 'd'}, inverse=True)
        assert f.serialize() == 'exclude,a,b,d'

    def test_parse(self):
        f = subview.MatchingUtteranceIdxFilter.parse('include,a,b,d')

        assert f.utterance_idxs == {'a', 'b', 'd'}
        assert not f.inverse

    def test_parse_inverse(self):
        f = subview.MatchingUtteranceIdxFilter.parse('exclude,a,b,d')

        assert f.utterance_idxs == {'a', 'b', 'd'}
        assert f.inverse


@pytest.fixture
def utt_without_noise():
    utt = tracks.Utterance('utt-1', 'file-1')

    utt.set_label_list(annotations.LabelList(idx='alpha', labels=[
        annotations.Label('music', 0, 5),
        annotations.Label('speech', 5, 12),
        annotations.Label('music', 13, 15)
    ]))

    utt.set_label_list(annotations.LabelList(idx='bravo', labels=[
        annotations.Label('music', 0, 1),
        annotations.Label('speech', 2, 6)
    ]))

    return utt


@pytest.fixture
def utt_with_noise():
    utt = tracks.Utterance('utt-2', 'file-2')

    utt.set_label_list(annotations.LabelList(idx='alpha', labels=[
        annotations.Label('music', 0, 5),
        annotations.Label('speech', 5, 12),
        annotations.Label('noise', 13, 15)
    ]))

    utt.set_label_list(annotations.LabelList(idx='bravo', labels=[
        annotations.Label('music', 0, 1),
        annotations.Label('speech', 2, 6)
    ]))

    return utt


class TestMatchingLabelFilter:

    def test_match_all_label_lists(self, utt_with_noise, utt_without_noise):
        filter = subview.MatchingLabelFilter(labels={'music', 'speech'})

        assert filter.match(utt_without_noise, None)
        assert not filter.match(utt_with_noise, None)

    def test_match_single(self, utt_with_noise, utt_without_noise):
        filter = subview.MatchingLabelFilter(labels={'music', 'speech'}, label_list_ids={'alpha'})

        assert filter.match(utt_without_noise, None)
        assert not filter.match(utt_with_noise, None)

        filter = subview.MatchingLabelFilter(labels={'music', 'speech'}, label_list_ids={'bravo'})

        assert filter.match(utt_without_noise, None)
        assert filter.match(utt_with_noise, None)

    def test_serialize(self):
        filter = subview.MatchingLabelFilter(labels={'music', 'speech'}, label_list_ids={'alpha'})

        assert filter.serialize() == 'alpha|||music,speech'

    def test_serialize_no_label_list_ids(self):
        filter = subview.MatchingLabelFilter(labels={'music', 'speech'})

        assert filter.serialize() == '|||music,speech'

    def test_parse(self):
        filter = subview.MatchingLabelFilter.parse('alpha|||music,speech')

        assert filter.labels == {'music', 'speech'}
        assert filter.label_list_ids == {'alpha'}

    def test_parse_no_label_list_ids(self):
        filter = subview.MatchingLabelFilter.parse('music,speech')

        assert filter.labels == {'music', 'speech'}
        assert filter.label_list_ids == set()


@pytest.fixture
def sample_subview():
        filter = subview.MatchingUtteranceIdxFilter(utterance_idxs={'utt-1', 'utt-3'})
        corpus = resources.create_dataset()
        return subview.Subview(corpus, filter_criteria=[filter])


class TestSubview:

    def test_tracks(self, sample_subview):
        assert sample_subview.num_tracks == 2
        assert 'wav-1' in sample_subview.tracks.keys()
        assert 'wav_3' in sample_subview.tracks.keys()

    def test_utterances(self, sample_subview):
        assert sample_subview.num_utterances == 2
        assert 'utt-1' in sample_subview.utterances.keys()
        assert 'utt-3' in sample_subview.utterances.keys()

    def test_issuers(self, sample_subview):
        assert sample_subview.num_issuers == 2
        assert 'spk-1' in sample_subview.issuers.keys()
        assert 'spk-2' in sample_subview.issuers.keys()

    def test_serialize(self, sample_subview):
        repr = sample_subview.serialize()

        assert repr == 'matching_utterance_ids\ninclude,utt-1,utt-3'

    def test_parse(self):
        corpus = resources.create_dataset()
        sv = subview.Subview.parse('matching_utterance_ids\ninclude,utt-1,utt-3', corpus=corpus)

        assert len(sv.filter_criteria) == 1
        assert sv.filter_criteria[0].utterance_idxs == {'utt-1', 'utt-3'}

    def test_utterances_without_issuers(self, sample_subview):
        sample_subview.corpus.utterances['utt-3'].issuer = None
        sample_subview.corpus.utterances['utt-4'].issuer = None
        sample_subview.corpus.utterances['utt-5'].issuer = None

        assert sample_subview.num_utterances == 2
        assert sample_subview.num_issuers == 1



from audiomate import corpus
from audiomate import tracks
from audiomate import annotations
from audiomate.corpus import validation

from tests import resources


class TestUtteranceTranscriptionRatioValidator:

    def test_name(self):
        val = validation.UtteranceTranscriptionRatioValidator(10, corpus.LL_WORD_TRANSCRIPT)

        assert val.name() == 'Utterance-Transcription-Ratio ({})'.format(corpus.LL_WORD_TRANSCRIPT)

    def test_validate(self):
        ds = resources.create_dataset()
        ds.utterances['utt-3'].set_label_list(annotations.LabelList.create_single(
            'max length here 11',
            idx=corpus.LL_WORD_TRANSCRIPT
        ))

        ds.utterances['utt-4'].set_label_list(annotations.LabelList.create_single(
            'too long here',
            idx=corpus.LL_WORD_TRANSCRIPT
        ))

        val = validation.UtteranceTranscriptionRatioValidator(10, corpus.LL_WORD_TRANSCRIPT)
        result = val.validate(ds)

        assert not result.passed
        assert len(result.invalid_items) == 1
        assert 'utt-4' in result.invalid_items.keys()


class TestLabelCountValidator:

    def test_name(self):
        val = validation.LabelCountValidator(1, corpus.LL_WORD_TRANSCRIPT)

        assert val.name() == 'Label-Count ({})'.format(corpus.LL_WORD_TRANSCRIPT)

    def test_validate(self):
        ds = resources.create_dataset()
        ds.utterances['utt-3'].set_label_list(annotations.LabelList(idx=corpus.LL_WORD_TRANSCRIPT))
        del ds.utterances['utt-4'].label_lists[corpus.LL_WORD_TRANSCRIPT]

        val = validation.LabelCountValidator(1, corpus.LL_WORD_TRANSCRIPT)
        result = val.validate(ds)

        assert not result.passed
        assert len(result.invalid_items) == 2
        assert result.invalid_items['utt-3'] == 'Only {} labels'.format(0)
        assert result.invalid_items['utt-4'] == 'No label-list {}'.format(corpus.LL_WORD_TRANSCRIPT)


class TestLabelCoverageValidator:

    def test_name(self):
        val = validation.LabelCoverageValidator('test_ll')

        assert val.name() == 'Label-Coverage (test_ll)'

    def test_validate_passes(self):
        ds = resources.create_single_label_corpus()

        val = validation.LabelCoverageValidator('default')
        result = val.validate(ds)

        assert result.passed
        assert len(result.uncovered_segments) == 0

    def test_validate(self):
        ds = resources.create_single_label_corpus()
        utt4_ll = annotations.LabelList(idx='default', labels=[
            annotations.Label('a', start=0.0, end=1.44),
            annotations.Label('a', start=1.89, end=10.0),
        ])
        ds.utterances['utt-4'].set_label_list(utt4_ll)
        utt6_ll = annotations.LabelList(idx='default', labels=[
            annotations.Label('a', start=1.33, end=5.9),
            annotations.Label('a', start=5.9, end=14.7),
        ])
        ds.utterances['utt-6'].set_label_list(utt6_ll)

        val = validation.LabelCoverageValidator('default')
        result = val.validate(ds)

        assert not result.passed
        assert set(result.uncovered_segments.keys()) == {'utt-4', 'utt-6'}

        assert result.uncovered_segments['utt-4'] == [(1.44, 1.89)]
        assert result.uncovered_segments['utt-6'] == [(0.0, 1.33), (14.7, 15.0)]


class TestLabelOverflowValidator:

    def test_name(self):
        val = validation.LabelOverflowValidator('test_ll')

        assert val.name() == 'Label-Overflow (test_ll)'

    def test_validate_passes(self):
        ds = resources.create_single_label_corpus()

        val = validation.LabelOverflowValidator('default')
        result = val.validate(ds)

        assert result.passed
        assert len(result.overflow_segments) == 0

    def test_validate_returns_part_of_overlapping_label(self):
        ds = resources.create_single_label_corpus()
        utt4_ll = annotations.LabelList(idx='default', labels=[
            annotations.Label('a', start=0.0, end=9.0),
            annotations.Label('b', start=9.0, end=13.0),
        ])
        ds.utterances['utt-4'].set_label_list(utt4_ll)
        utt6_ll = annotations.LabelList(idx='default', labels=[
            annotations.Label('a', start=-2.0, end=5.9),
            annotations.Label('b', start=5.9, end=14.7),
        ])
        ds.utterances['utt-6'].set_label_list(utt6_ll)

        val = validation.LabelOverflowValidator('default')
        result = val.validate(ds)

        assert not result.passed
        assert set(result.overflow_segments.keys()) == {'utt-4', 'utt-6'}

        assert result.overflow_segments['utt-4'] == [(10.0, 13.0, 'b')]
        assert result.overflow_segments['utt-6'] == [(-2.0, 0.0, 'a')]

    def test_validate_utterance_returns_completly_outlying_label(self):
        utt = tracks.Utterance('utt-idx', None, start=10.0, end=17.9)
        ll = annotations.LabelList(idx='default', labels=[
            annotations.Label('a', start=-4.0, end=-2.0),
            annotations.Label('b', start=19.0, end=22.0),
        ])
        utt.set_label_list(ll)

        val = validation.LabelOverflowValidator('default')
        result = val.validate_utterance(utt)
        result = sorted(result, key=lambda x: x[0])

        assert len(result) == 2

        assert result[0] == (-4.0, -2.0, 'a')
        assert result[1] == (19.0, 22.0, 'b')



import os
import glob

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.utils import jsonfile

from . import base
from . import downloader

MASTER_DOWNLOAD_URL = 'https://github.com/soerenab/AudioMNIST/archive/master.zip'


class AudioMNISTDownloader(downloader.ArchiveDownloader):
    """
    Downloader for the audioMNIST dataset.

    Args:
        url (str): The url to download the dataset from. If not given the default URL is used.
                   It is expected to be a zip file.
    """

    def __init__(self, url=None):
        if url is None:
            url = MASTER_DOWNLOAD_URL

        super(AudioMNISTDownloader, self).__init__(
            url,
            move_files_up=True
        )

    @classmethod
    def type(cls):
        return 'audio-mnist'


class AudioMNISTReader(base.CorpusReader):
    """
    Reader for the audioMNIST Corpus.

    .. seealso::

       `AudioMNIST-Dataset <https://github.com/soerenab/AudioMNIST>`_
          Download page
    """

    @classmethod
    def type(cls):
        return 'audio-mnist'

    def _check_for_missing_files(self, path):
        missing_files = []

        recordings_folder = os.path.join(path, 'data')
        if not os.path.isdir(recordings_folder):
            missing_files.append(recordings_folder)

        meta_file = os.path.join(path, 'data', 'audioMNIST_meta.txt')
        if not os.path.isfile(meta_file):
            missing_files.append(meta_file)

        return missing_files

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)
        data_path = os.path.join(path, 'data')
        meta_data = AudioMNISTReader.load_speaker_meta(path)

        for speaker_idx in os.listdir(data_path):
            speaker_path = os.path.join(data_path, speaker_idx)

            if os.path.isdir(speaker_path):

                for file_path in glob.glob(os.path.join(speaker_path, '*.wav')):
                    file_idx = os.path.splitext(os.path.basename(file_path))[0]

                    corpus.new_file(file_path, file_idx)

                    idx_parts = file_idx.split('_')
                    digit = idx_parts[0]

                    if speaker_idx not in corpus.issuers.keys():
                        issuer = issuers.Speaker(
                            speaker_idx,
                            gender=AudioMNISTReader.get_gender(meta_data, speaker_idx),
                            age_group=AudioMNISTReader.get_age_group(meta_data, speaker_idx)
                        )
                        corpus.import_issuers(issuer)

                    utterance = corpus.new_utterance(file_idx, file_idx, speaker_idx)
                    utterance.set_label_list(annotations.LabelList.create_single(
                        str(digit),
                        idx=audiomate.corpus.LL_WORD_TRANSCRIPT
                    ))

        return corpus

    @staticmethod
    def load_speaker_meta(corpus_path):
        meta_file = os.path.join(corpus_path, 'data', 'audioMNIST_meta.txt')
        return jsonfile.read_json_file(meta_file)

    @staticmethod
    def get_gender(meta_data, speaker_idx):
        gender_str = meta_data[speaker_idx]['gender']

        if gender_str == 'male':
            return issuers.Gender.MALE
        elif gender_str == 'female':
            return issuers.Gender.FEMALE
        else:
            return issuers.Gender.UNKNOWN

    @staticmethod
    def get_age_group(meta_data, speaker_idx):
        age_str = int(meta_data[speaker_idx]['age'])

        if age_str < 12:
            return issuers.AgeGroup.CHILD
        elif age_str < 18:
            return issuers.AgeGroup.YOUTH
        elif age_str < 65:
            return issuers.AgeGroup.ADULT
        else:
            return issuers.AgeGroup.SENIOR



import random
import os

import audiomate
from audiomate import issuers
from audiomate import tracks
from audiomate import annotations


def get_test_resource_path(sub_path_components):
    """ Get the absolute path of a file in the resources folder with its relative path components. """
    this_folder = os.path.dirname(__file__)
    root_folder = os.path.join(this_folder, '..', '..')
    return os.path.abspath(os.path.join(root_folder, 'tests', 'resources', *sub_path_components))


def sample_corpus_path(name):
    """ Return the path to a sample corpus path with its name. """
    return get_test_resource_path(['sample_corpora', name])


def generate_corpus(n_issuers,
                    n_tracks_per_issuer,
                    n_utts_per_track,
                    n_ll_per_utt,
                    n_label_per_ll,
                    rand=None):
    """
    Generate a corpus with mock data.
    """
    corpus = audiomate.Corpus()

    for issuer in generate_issuers(n_issuers, rand):
        corpus.import_issuers(issuer)

        n_tracks = rand.randint(*n_tracks_per_issuer)
        tracks = generate_tracks(n_tracks, rand)
        corpus.import_tracks(tracks)

        n_utts = rand.randint(*n_utts_per_track)
        for track in tracks:
            utts = generate_utterances(
                track,
                issuer,
                n_utts,
                n_ll_per_utt,
                n_label_per_ll,
                rand
            )

            corpus.import_utterances(utts)

    return corpus


def generate_issuers(n, rand=None):
    if rand is None:
        rand = random.Random()

    items = []

    for issuer_index in range(n):
        issuer_idx = 'issuer-{}'.format(issuer_index)

        issuer_type = rand.randint(1, 3)

        if issuer_type == 1:
            issuer = issuers.Speaker(
                issuer_idx,
                gender=issuers.Gender.UNKNOWN,
                age_group=issuers.AgeGroup.CHILD,
                native_language='de'
            )
        elif issuer_type == 2:
            issuer = issuers.Artist(issuer_idx, 'badam')
        else:
            issuer = issuers.Issuer(issuer_idx)

        items.append(issuer)

    return items


def generate_tracks(n, rand=None):
    if rand is None:
        rand = random.Random()

    items = []

    for i in range(n):
        track_idx = 'track-{}'.format(i)
        path = '/fake/{}.wav'.format(track_idx)
        track = tracks.FileTrack(track_idx, path)

        items.append(track)

    return items


def generate_utterances(track, issuer, n, n_ll_range, n_label_range, rand=None):
    if rand is None:
        rand = random.Random()

    items = []

    for i in range(n):
        utt_idx = '{}-utt-{}'.format(track.idx, i)
        start = rand.random() * 3
        end = 3 + rand.random() * 8

        utt = tracks.Utterance(
            utt_idx,
            track,
            issuer=issuer,
            start=start,
            end=end
        )
        n_ll = rand.randint(*n_ll_range)

        for ll in generate_label_lists(n_ll, n_label_range, rand=rand):
            utt.set_label_list(ll)

        items.append(utt)

    return items


def generate_label_lists(n, n_label_range, rand=None):
    if rand is None:
        rand = random.Random()

    items = []

    for i in range(n):
        n_labels = rand.randint(*n_label_range)
        ll_idx = 'll-{}'.format(i)
        ll = annotations.LabelList(
            idx=ll_idx,
            labels=generate_labels(n_labels)
        )

        items.append(ll)

    return items


def generate_labels(n):
    items = []

    for i in range(n):
        label = annotations.Label('label-{}'.format(i))
        items.append(label)

    return items



import abc


class ValidationResult:
    """
    Representation of the result of a validation.
    The basic result just indicates a pass or fail.
    Depending on the validator it can be extended to hold more information
    (e.g. utterance-ids which triggered the task to fail).

    Args:
        passed (bool): A boolean indicating, if the validation has passed (True) or failed (False).
        name (str): The name of the validator, that produced the result.
        info (dict): Dictionary containing key/value string-pairs with detailed information of the validation.
                     For example id of the label-list that was validated.
    """

    def __init__(self, passed, name='Validation', info=None):
        self.passed = passed
        self.name = name
        self.info = info

    def get_report(self):
        """
        Return a string containing a report of the result.
        This can used to print or save to a text file.

        Returns:
            str: String containing infos about the result
        """

        lines = [
            self.name,
            '=' * len(self.name)
        ]

        if self.info is not None:
            lines.append('')
            sorted_info = sorted(self.info.items(), key=lambda x: x[0])
            lines.extend(['--> {}: {}'.format(k, v) for k, v in sorted_info])

        lines.append('')
        lines.append('Result: {}'.format('Passed' if self.passed else 'Failed'))

        return '\n'.join(lines)


class InvalidItemsResult(ValidationResult):
    """
    A generic result class for validators that return a list of items (utterances, tracks) that were classified invalid.
    Besides the id of the item, a reason may be appended.

    Args:
        passed (bool): A boolean indicating, if the validation has passed (True) or failed (False).
        invalid_items (dict): A dictionary containing item-ids, that are invalid.
                                   The values are reasons why they are invalid.
        name (str): The name of the validator, that produced the result.
        info (dict): Dictionary containing key/value string-pairs with detailed information of the validation.
                     For example id of the label-list that was validated.
    """

    def __init__(self, passed, invalid_items, name='Validation', item_name='Utterances', info=None):
        super(InvalidItemsResult, self).__init__(passed, name=name, info=info)

        self.invalid_items = invalid_items
        self.item_name = item_name

    def get_report(self):
        """
        Return a string containing a report of the result.
        This can used to print or save to a text file.

        Returns:
            str: String containing infos about the result
        """

        lines = [super(InvalidItemsResult, self).get_report()]

        if len(self.invalid_items) > 0:
            lines.append('\nInvalid {}:'.format(self.item_name))

            sorted_items = sorted(self.invalid_items.items(), key=lambda x: x[0])
            lines.extend(['    * {} ({})'.format(x, y) for x, y in sorted_items])

        return '\n'.join(lines)


class Validator(abc.ABC):
    """
    A validator is a class that tests a specific behaviour/state of a corpus.
    """

    @abc.abstractmethod
    def name(self):
        """ Return a name, identifying the task. """
        pass

    @abc.abstractmethod
    def validate(self, corpus):
        """
        Perform the validation on the given corpus.

        Args:
            corpus (Corpus): The corpus to test/validate.

        Returns:
            ValidationResult: The result containing at least the pass/fail indication.
        """
        pass



import os

import pytest
import requests_mock

from audiomate import corpus
from audiomate.corpus import io
from audiomate.corpus.io import gtzan

from tests import resources
from . import reader_test as rt


@pytest.fixture()
def tar_data():
    with open(resources.get_resource_path(['sample_files', 'cv_corpus_v1.tar.gz']), 'rb') as f:
        return f.read()


class TestGtzanDownloader:

    def test_download(self, tar_data, tmpdir):
        target_folder = tmpdir.strpath
        downloader = io.GtzanDownloader()

        with requests_mock.Mocker() as mock:
            mock.get(gtzan.DOWNLOAD_URL, content=tar_data)

            downloader.download(target_folder)

        assert os.path.isfile(os.path.join(target_folder, 'cv-valid-dev.csv'))
        assert os.path.isdir(os.path.join(target_folder, 'cv-valid-dev'))
        assert os.path.isfile(os.path.join(target_folder, 'cv-valid-train.csv'))
        assert os.path.isdir(os.path.join(target_folder, 'cv-valid-train'))


class TestGtzanReader(rt.CorpusReaderTest):

    SAMPLE_PATH = resources.sample_corpus_path('gtzan')
    FILE_TRACK_BASE_PATH = SAMPLE_PATH

    EXPECTED_NUMBER_OF_TRACKS = 4
    EXPECTED_TRACKS = [
        rt.ExpFileTrack('bagpipe', os.path.join('music_wav', 'bagpipe.wav')),
        rt.ExpFileTrack('ballad', os.path.join('music_wav', 'ballad.wav')),
        rt.ExpFileTrack('acomic', os.path.join('speech_wav', 'acomic.wav')),
        rt.ExpFileTrack('acomic2', os.path.join('speech_wav', 'acomic2.wav')),
    ]

    EXPECTED_NUMBER_OF_ISSUERS = 0

    EXPECTED_NUMBER_OF_UTTERANCES = 4
    EXPECTED_UTTERANCES = [
        rt.ExpUtterance('bagpipe', 'bagpipe', None, 0, float('inf')),
        rt.ExpUtterance('ballad', 'ballad', None, 0, float('inf')),
        rt.ExpUtterance('acomic', 'acomic', None, 0, float('inf')),
        rt.ExpUtterance('acomic2', 'acomic2', None, 0, float('inf')),
    ]

    EXPECTED_LABEL_LISTS = {
        'ballad': [
            rt.ExpLabelList(corpus.LL_DOMAIN, 1),
        ],
        'bagpipe': [
            rt.ExpLabelList(corpus.LL_DOMAIN, 1),
        ],
        'acomic': [
            rt.ExpLabelList(corpus.LL_DOMAIN, 1),
        ],
        'acomic2': [
            rt.ExpLabelList(corpus.LL_DOMAIN, 1),
        ],
    }

    EXPECTED_LABELS = {
        'bagpipe': [
            rt.ExpLabel(corpus.LL_DOMAIN, corpus.LL_DOMAIN_MUSIC, 0, float('inf')),
        ],
        'ballad': [
            rt.ExpLabel(corpus.LL_DOMAIN, corpus.LL_DOMAIN_MUSIC, 0, float('inf')),
        ],
        'acomic': [
            rt.ExpLabel(corpus.LL_DOMAIN, corpus.LL_DOMAIN_SPEECH, 0, float('inf')),
        ],
        'acomic2': [
            rt.ExpLabel(corpus.LL_DOMAIN, corpus.LL_DOMAIN_SPEECH, 0, float('inf')),
        ],
    }

    EXPECTED_NUMBER_OF_SUBVIEWS = 0

    def load(self):
        return io.GtzanReader().load(self.SAMPLE_PATH)



import random
import collections

from audiomate.corpus.subset import subview
from audiomate.corpus.subset import utils


class Splitter(object):
    """
    A splitter provides methods for splitting a corpus into different subsets.
    It provides different approaches for splitting the corpus.
    (Methods indicated by ``split_by_``)
    These methods mostly take some proportions parameter,
    which defines how big (in relation) the
    subsets should be. The subsets are returned
    as :py:class:`audiomate.corpus.Subview`.

    Args:
        corpus (Corpus): The corpus that should be splitted.
        random_seed (int): Seed to use for random number generation.
    """

    def __init__(self, corpus, random_seed=None):
        self.corpus = corpus
        self.rand = random.Random()
        self.rand.seed(a=random_seed)

    def split(self, proportions={}, separate_issuers=False):
        """
        Split the corpus based on the number of utterances.
        The utterances are distributed to `len(proportions)` subsets,
        according to the ratios `proportions[subset]`.

        Args:
            proportions (dict): A dictionary containing the relative size of
                                the target subsets. The key is an identifier
                                for the subset.
            separate_issuers (bool): If True it makes sure that all utterances
                                     of an issuer are in the same subset.

        Returns:
            (dict): A dictionary containing the subsets with the identifier
                    from the input as key.

        Example::

            >>> spl = Splitter(corpus)
            >>> corpus.num_utterances
            100
            >>> subsets = spl.split(proportions={
            >>>     "train" : 0.6,
            >>>     "dev" : 0.2,
            >>>     "test" : 0.2
            >>> })
            >>> print(subsets)
            {'dev': <audiomate.corpus.subview.Subview at 0x104ce7400>,
            'test': <audiomate.corpus.subview.Subview at 0x104ce74e0>,
            'train': <audiomate.corpus.subview.Subview at 0x104ce7438>}
            >>> subsets['train'].num_utterances
            60
            >>> subsets['dev'].num_utterances
            20
            >>> subsets['test'].num_utterances
            20
        """
        if separate_issuers:
            # Count number of utterances per issuer
            issuer_utt_count = collections.defaultdict(int)
            issuer_utts = collections.defaultdict(list)

            for utterance in self.corpus.utterances.values():
                issuer_utt_count[utterance.issuer.idx] += 1
                issuer_utts[utterance.issuer.idx].append(utterance.idx)

            issuer_utt_count = {
                k: {'count': int(v)}
                for k, v in issuer_utt_count.items()
            }

            # Split with total utt count per issuer as weight
            issuer_splits = utils.get_identifiers_splitted_by_weights(
                issuer_utt_count,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_issuer_splits(issuer_splits, issuer_utts)

        else:
            utterance_idxs = list(self.corpus.utterances.keys())
            splits = utils.split_identifiers(
                utterance_idxs,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_utterance_splits(splits)

    def split_by_audio_duration(self, proportions={}, separate_issuers=False):
        """
        Split the corpus based on the the total duration of audio.
        The utterances are distributed to `len(proportions)` subsets.
        Utterances are split up in a way that each subset contains
        audio with a duration proportional to the given proportions.

        Args:
            proportions (dict): A dictionary containing the relative size of
                                the target subsets. The key is an identifier
                                for the subset.
            separate_issuers (bool): If True it makes sure that all utterances
                                     of an issuer are in the same subset.

        Returns:
            (dict): A dictionary containing the subsets with the identifier
                    from the input as key.

        Example::

            >>> spl = Splitter(corpus)
            >>> corpus.num_utterances
            100
            >>> subsets = spl.split_by_audio_duration(proportions={
            >>>     "train" : 0.6,
            >>>     "dev" : 0.2,
            >>>     "test" : 0.2
            >>> })
            >>> print(subsets)
            {'dev': <audiomate.corpus.subview.Subview at 0x104ce7400>,
            'test': <audiomate.corpus.subview.Subview at 0x104ce74e0>,
            'train': <audiomate.corpus.subview.Subview at 0x104ce7438>}
            >>> subsets['train'].num_utterances
            55
            >>> subsets['dev'].num_utterances
            35
            >>> subsets['test'].num_utterances
            10
        """

        if separate_issuers:
            # Count total length of utterances per issuer
            issuer_utts_duration = collections.defaultdict(float)
            issuer_utts = collections.defaultdict(list)

            for utterance in self.corpus.utterances.values():
                issuer_utts_duration[utterance.issuer.idx] += utterance.duration
                issuer_utts[utterance.issuer.idx].append(utterance.idx)

            issuer_utts_duration = {
                k: {'duration': int(v * 1000)}
                for k, v in issuer_utts_duration.items()
            }

            # Split with total utt duration per issuer as weight
            issuer_splits = utils.get_identifiers_splitted_by_weights(
                issuer_utts_duration,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_issuer_splits(issuer_splits, issuer_utts)

        else:
            utterance_to_duration = {}

            for utterance in self.corpus.utterances.values():
                utterance_to_duration[utterance.idx] = {
                    'duration': int(utterance.duration * 1000)
                }

            splits = utils.get_identifiers_splitted_by_weights(
                utterance_to_duration,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_utterance_splits(splits)

    def split_by_label_length(self, proportions={},
                              label_list_idx=None,
                              separate_issuers=False):
        """
        Split the corpus based on the the total length of the label-list.
        The utterances are distributed to `len(proportions)` subsets.
        Utterances are split up in a way that each subset contains
        labels summed up to a length proportional to the given proportions.
        Length is defined as the number of characters.

        Args:
            proportions (dict): A dictionary containing the relative size
                                of the target subsets.
                                The key is an identifier for the subset.
            label_list_idx (str): The idx of the label-list to use for compute
                                  the length. If `None` all label-lists are used.
            separate_issuers (bool): If True it makes sure that all utterances
                                     of an issuer are in the same subset.

        Returns:
            (dict): A dictionary containing the subsets with the identifier
                    from the input as key.

        Example::

            >>> spl = Splitter(corpus)
            >>> corpus.num_utterances
            100
            >>> subsets = spl.split_by_label_length(proportions={
            >>>     "train" : 0.6,
            >>>     "dev" : 0.2,
            >>>     "test" : 0.2
            >>> })
            >>> print(subsets)
            {'dev': <audiomate.corpus.subview.Subview at 0x104ce7400>,
            'test': <audiomate.corpus.subview.Subview at 0x104ce74e0>,
            'train': <audiomate.corpus.subview.Subview at 0x104ce7438>}
            >>> subsets['train'].num_utterances
            55
            >>> subsets['dev'].num_utterances
            35
            >>> subsets['test'].num_utterances
            10
        """

        if separate_issuers:
            # Count total length of utterances per issuer
            issuer_utts_length = collections.defaultdict(int)
            issuer_utts = collections.defaultdict(list)

            for utterance in self.corpus.utterances.values():
                lls = utterance.label_lists

                if label_list_idx is None:
                    num_char = sum([ll.total_length for ll in lls.values()])
                else:
                    num_char = lls[label_list_idx].total_length

                issuer_utts_length[utterance.issuer.idx] += num_char
                issuer_utts[utterance.issuer.idx].append(utterance.idx)

            issuer_utts_length = {
                k: {'length': v}
                for k, v in issuer_utts_length.items()
            }

            # Split with total utt duration per issuer as weight
            issuer_splits = utils.get_identifiers_splitted_by_weights(
                issuer_utts_length,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_issuer_splits(issuer_splits, issuer_utts)

        else:
            utterance_to_length = {}

            for utterance in self.corpus.utterances.values():
                lls = utterance.label_lists

                if label_list_idx is None:
                    num_char = sum([ll.total_length for ll in lls.values()])
                else:
                    num_char = lls[label_list_idx].total_length

                utterance_to_length[utterance.idx] = {
                    'length': num_char
                }

            splits = utils.get_identifiers_splitted_by_weights(
                utterance_to_length,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_utterance_splits(splits)

    def split_by_label_occurence(self, proportions={}, separate_issuers=False):
        """
        Split the corpus based on the total number of occcurences of labels.
        The utterances are distributed to `len(proportions)` subsets.
        Utterances are split up in a way that each subset contains
        labels-occurences proportional to the given proportions.

        Args:
            proportions (dict): A dictionary containing the relative size
                                of the target subsets.
                                The key is an identifier for the subset.
            separate_issuers (bool): If True it makes sure that all utterances
                                     of an issuer are in the same subset.

        Returns:
            (dict): A dictionary containing the subsets with the identifier
                    from the input as key.

        Example::

            >>> spl = Splitter(corpus)
            >>> corpus.num_utterances
            100
            >>> subsets = spl.split_by_label_occurence(proportions={
            >>>     "train" : 0.6,
            >>>     "dev" : 0.2,
            >>>     "test" : 0.2
            >>> })
            >>> print(subsets)
            {'dev': <audiomate.corpus.subview.Subview at 0x104ce7400>,
            'test': <audiomate.corpus.subview.Subview at 0x104ce74e0>,
            'train': <audiomate.corpus.subview.Subview at 0x104ce7438>}
            >>> subsets['train'].num_utterances
            55
            >>> subsets['dev'].num_utterances
            35
            >>> subsets['test'].num_utterances
            10
        """

        if separate_issuers:
            # Count total length of utterances per issuer
            issuer_label_count = collections.defaultdict(collections.Counter)
            issuer_utts = collections.defaultdict(list)

            for utterance in self.corpus.utterances.values():
                label_count = utterance.label_count()
                issuer_label_count[utterance.issuer.idx].update(label_count)
                issuer_utts[utterance.issuer.idx].append(utterance.idx)

            # issuer_label_count = {
            #     k: dict(v) for k, v in issuer_label_count.items()
            # }

            # Split with total utt duration per issuer as weight
            issuer_splits = utils.get_identifiers_splitted_by_weights(
                issuer_label_count,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_issuer_splits(issuer_splits, issuer_utts)

        else:
            utterance_label_count = {
                utt.idx: dict(utt.label_count())
                for utt in self.corpus.utterances.values()
            }

            splits = utils.get_identifiers_splitted_by_weights(
                utterance_label_count,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_utterance_splits(splits)

    def split_by_label_duration(self, proportions={}, separate_issuers=False):
        """
        Split the corpus based on the total duration of labels (end - start).
        The utterances are distributed to `len(proportions)` subsets.
        Utterances are split up in a way that each subset contains
        labels with a duration proportional to the given proportions.

        Args:
            proportions (dict): A dictionary containing the relative size of
                                the target subsets. The key is an identifier
                                for the subset.
            separate_issuers (bool): If True it makes sure that all utterances
                                     of an issuer are in the same subset.

        Returns:
            (dict): A dictionary containing the subsets with the identifier
                    from the input as key.

        Example::

            >>> spl = Splitter(corpus)
            >>> corpus.num_utterances
            100
            >>> subsets = spl.split_by_label_duration(proportions={
            >>>     "train" : 0.6,
            >>>     "dev" : 0.2,
            >>>     "test" : 0.2
            >>> })
            >>> print(subsets)
            {'dev': <audiomate.corpus.subview.Subview at 0x104ce7400>,
            'test': <audiomate.corpus.subview.Subview at 0x104ce74e0>,
            'train': <audiomate.corpus.subview.Subview at 0x104ce7438>}
            >>> subsets['train'].num_utterances
            55
            >>> subsets['dev'].num_utterances
            35
            >>> subsets['test'].num_utterances
            10
        """

        if separate_issuers:
            # Count total length of utterances per issuer
            issuer_label_duration = collections.defaultdict(collections.Counter)
            issuer_utts = collections.defaultdict(list)

            for utterance in self.corpus.utterances.values():
                issuer_label_duration[utterance.issuer.idx].update(
                    utterance.label_total_duration()
                )
                issuer_utts[utterance.issuer.idx].append(utterance.idx)

            # Split with total utt duration per issuer as weight
            issuer_splits = utils.get_identifiers_splitted_by_weights(
                issuer_label_duration,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_issuer_splits(issuer_splits, issuer_utts)

        else:
            utterance_label_duration = {
                utt.idx: dict(utt.label_total_duration())
                for utt in self.corpus.utterances.values()
            }

            splits = utils.get_identifiers_splitted_by_weights(
                utterance_label_duration,
                proportions,
                seed=self.rand.random()
            )

            return self._subviews_from_utterance_splits(splits)

    def _subviews_from_utterance_splits(self, splits):
        """
        Create subviews from a dict containing utterance-ids for each subview.

        e.g. {'train': ['utt-1', 'utt-2'], 'test': [...], ...}
        """
        subviews = {}

        for idx, subview_utterances in splits.items():
            filter = subview.MatchingUtteranceIdxFilter(utterance_idxs=subview_utterances)
            split = subview.Subview(self.corpus, filter_criteria=filter)
            subviews[idx] = split

        return subviews

    def _subviews_from_issuer_splits(self, splits, issuer_to_utt):
        """
        Create subviews from a dict containing issuer-ids for each subview
        and a map from issuer-ids to utterance-ids.

        e.g. {'train': ['issuer-1', 'issuer-2'], 'test': [...], ...}
        """

        # Collect utterances of all issuers per split
        utt_splits = collections.defaultdict(list)

        for split_idx, issuer_ids in splits.items():
            for issuer_idx in issuer_ids:
                utt_splits[split_idx].extend(issuer_to_utt[issuer_idx])

        return self._subviews_from_utterance_splits(utt_splits)



import os
import tempfile
import pytest

from audiomate import corpus
from audiomate.corpus import io
from audiomate.utils import textfile

from tests import resources


@pytest.fixture()
def writer():
    return io.Wav2LetterWriter()


@pytest.fixture()
def path():
    return tempfile.mkdtemp()


class TestWav2LetterWriter:

    def test_save_all(self, writer, tmpdir):
        ds = resources.create_dataset()
        writer.save(ds, tmpdir.strpath)

        all_path = os.path.join(tmpdir.strpath, 'all.lst')

        assert os.path.isfile(all_path)

        records = textfile.read_separated_lines(all_path, separator=' ', max_columns=4)

        assert len(records) == 5

        # DATA RECORDS
        utts = {r[0]: (r[1], r[2], r[3]) for r in records}

        utt_idx = 'utt-1'
        assert utts[utt_idx][0] == ds.utterances[utt_idx].track.path
        assert utts[utt_idx][1] == '41523'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        utt_idx = 'utt-2'
        assert utts[utt_idx][0] == ds.utterances[utt_idx].track.path
        assert utts[utt_idx][1] == '41523'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        utt_idx = 'utt-3'
        assert utts[utt_idx][0] == os.path.join(tmpdir.strpath, 'audio', '{}.wav'.format(utt_idx))
        assert utts[utt_idx][1] == '24000'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        utt_idx = 'utt-4'
        assert utts[utt_idx][0] == os.path.join(tmpdir.strpath, 'audio', '{}.wav'.format(utt_idx))
        assert utts[utt_idx][1] == '16000'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        utt_idx = 'utt-5'
        assert utts[utt_idx][0] == ds.utterances[utt_idx].track.path
        assert utts[utt_idx][1] == '41523'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

    def test_save_subset_train(self, writer, tmpdir):
        ds = resources.create_dataset()
        writer.save(ds, tmpdir.strpath)

        all_path = os.path.join(tmpdir.strpath, 'train.lst')

        assert os.path.isfile(all_path)

        records = textfile.read_separated_lines(all_path, separator=' ', max_columns=4)

        assert len(records) == 3

        utts = {r[0]: (r[1], r[2], r[3]) for r in records}

        utt_idx = 'utt-1'
        assert utts[utt_idx][0] == ds.utterances[utt_idx].track.path
        assert utts[utt_idx][1] == '41523'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        utt_idx = 'utt-2'
        assert utts[utt_idx][0] == ds.utterances[utt_idx].track.path
        assert utts[utt_idx][1] == '41523'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        utt_idx = 'utt-3'
        assert utts[utt_idx][0] == os.path.join(tmpdir.strpath, 'audio', '{}.wav'.format(utt_idx))
        assert utts[utt_idx][1] == '24000'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

    def test_save_subset_dev(self, writer, tmpdir):
        ds = resources.create_dataset()
        writer.save(ds, tmpdir.strpath)

        all_path = os.path.join(tmpdir.strpath, 'dev.lst')

        assert os.path.isfile(all_path)

        records = textfile.read_separated_lines(all_path, separator=' ', max_columns=4)

        assert len(records) == 2

        # DATA RECORDS
        utts = {r[0]: (r[1], r[2], r[3]) for r in records}

        utt_idx = 'utt-4'
        assert utts[utt_idx][0] == os.path.join(tmpdir.strpath, 'audio', '{}.wav'.format(utt_idx))
        assert utts[utt_idx][1] == '16000'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value

        utt_idx = 'utt-5'
        assert utts[utt_idx][0] == ds.utterances[utt_idx].track.path
        assert utts[utt_idx][1] == '41523'
        assert utts[utt_idx][2] == ds.utterances[utt_idx].label_lists[corpus.LL_WORD_TRANSCRIPT].labels[0].value



import os

import audiomate
from audiomate import annotations
from audiomate import issuers
from audiomate.corpus import subset
from audiomate.utils import download
from audiomate.utils import textfile
from . import base

DOWNLOAD_URLS = {
    'de_DE': 'http://www.m-ailabs.bayern/?ddownload=410',
    'en_UK': 'http://www.m-ailabs.bayern/?ddownload=412',
    'en_US': 'http://www.m-ailabs.bayern/?ddownload=411',
    'es_ES': 'http://www.m-ailabs.bayern/?ddownload=413',
    'it_IT': 'http://www.m-ailabs.bayern/?ddownload=409',
    'uk_UK': 'http://www.m-ailabs.bayern/?ddownload=414',
    'ru_RU': 'http://www.m-ailabs.bayern/?ddownload=415',
    'fr_FR': 'http://www.m-ailabs.bayern/?ddownload=681',
    'pl_PL': 'http://www.m-ailabs.bayern/?ddownload=673',
}


class MailabsDownloader(base.CorpusDownloader):
    """
    Downloader for the M-AILABS Speech Dataset.

    Args:
        tags (list): List of tags for different parts to download.
                     Corresponds to the tags in the
                     `Statistics & Download Links` on the webpage.
                     If ``None``, all parts are downloaded.
    """

    def __init__(self, tags=None):
        self.tags = tags

    @classmethod
    def type(cls):
        return 'mailabs'

    def _download(self, target_path):
        os.makedirs(target_path, exist_ok=True)

        for tag, download_url in DOWNLOAD_URLS.items():
            if self.tags is None or tag in self.tags:
                tmp_file = os.path.join(target_path, 'tmp_{}.tgz'.format(tag))

                download.download_file(download_url, tmp_file)
                download.extract_tar(tmp_file, target_path)

                os.remove(tmp_file)


class MailabsReader(base.CorpusReader):
    """
    Reader for the M-AILABS Speech Dataset.

    .. seealso::

       `M-AILABS Speech Dataset <http://www.m-ailabs.bayern/en/the-mailabs-speech-dataset/>`_
          Project Page
    """

    @classmethod
    def type(cls):
        return 'mailabs'

    def _check_for_missing_files(self, path):
        return []

    def _load(self, path):
        corpus = audiomate.Corpus(path=path)
        tag_folders = MailabsReader.get_folders(path)

        for tag_folder in tag_folders:
            MailabsReader.load_tag(corpus, tag_folder)

        return corpus

    @staticmethod
    def get_folders(path):
        """
        Return a list of all subfolder-paths in the given path.
        """
        folder_paths = []

        for item in os.listdir(path):
            folder_path = os.path.join(path, item)

            if os.path.isdir(folder_path):
                folder_paths.append(folder_path)

        return folder_paths

    @staticmethod
    def load_tag(corpus, path):
        """
        Iterate over all speakers on load them.
        Collect all utterance-idx and create a subset of them.
        """
        tag_idx = os.path.basename(path)
        data_path = os.path.join(path, 'by_book')
        tag_utt_ids = []

        for gender_path in MailabsReader.get_folders(data_path):
            # IN MIX FOLDERS THERE ARE NO SPEAKERS
            # HANDLE EVERY UTT AS DIFFERENT ISSUER
            if os.path.basename(gender_path) == 'mix':
                utt_ids = MailabsReader.load_books_of_speaker(corpus,
                                                              gender_path,
                                                              None)

                tag_utt_ids.extend(utt_ids)

            else:
                for speaker_path in MailabsReader.get_folders(gender_path):
                    speaker = MailabsReader.load_speaker(corpus, speaker_path)
                    utt_ids = MailabsReader.load_books_of_speaker(corpus,
                                                                  speaker_path,
                                                                  speaker)

                    tag_utt_ids.extend(utt_ids)

        filter = subset.MatchingUtteranceIdxFilter(
            utterance_idxs=set(tag_utt_ids)
        )
        subview = subset.Subview(corpus, filter_criteria=[filter])
        corpus.import_subview(tag_idx, subview)

    @staticmethod
    def load_speaker(corpus, path):
        """
        Create a speaker instance for the given path.
        """
        base_path, speaker_name = os.path.split(path)
        base_path, gender_desc = os.path.split(base_path)
        base_path, __ = os.path.split(base_path)
        base_path, tag = os.path.split(base_path)

        gender = issuers.Gender.UNKNOWN

        if gender_desc == 'male':
            gender = issuers.Gender.MALE
        elif gender_desc == 'female':
            gender = issuers.Gender.FEMALE

        speaker = issuers.Speaker(speaker_name, gender=gender)
        corpus.import_issuers(speaker)

        return speaker

    @staticmethod
    def load_books_of_speaker(corpus, path, speaker):
        """
        Load all utterances for the speaker at the given path.
        """
        utt_ids = []

        for book_path in MailabsReader.get_folders(path):
            meta_path = os.path.join(book_path, 'metadata.csv')
            wavs_path = os.path.join(book_path, 'wavs')

            meta = textfile.read_separated_lines(meta_path,
                                                 separator='|',
                                                 max_columns=3)

            for entry in meta:
                file_basename = entry[0]
                transcription_raw = entry[1]
                transcription_clean = entry[2]

                if speaker is None:
                    idx = file_basename
                    utt_speaker = issuers.Speaker(idx)
                    speaker_idx = idx
                    corpus.import_issuers(utt_speaker)
                else:
                    idx = '{}-{}'.format(speaker.idx, file_basename)
                    speaker_idx = speaker.idx

                wav_name = '{}.wav'.format(file_basename)
                wav_path = os.path.join(wavs_path, wav_name)

                if os.path.isfile(wav_path):
                    corpus.new_file(wav_path, idx)

                    ll_raw = annotations.LabelList.create_single(
                        transcription_raw,
                        idx=audiomate.corpus.LL_WORD_TRANSCRIPT_RAW
                    )

                    ll_clean = annotations.LabelList.create_single(
                        transcription_clean,
                        idx=audiomate.corpus.LL_WORD_TRANSCRIPT
                    )

                    utterance = corpus.new_utterance(idx, idx, speaker_idx)
                    utterance.set_label_list(ll_raw)
                    utterance.set_label_list(ll_clean)

                    utt_ids.append(utterance.idx)

        return utt_ids



"""
This module contains classes to represent issuers fo audio samples.
An issuer represents the person/machine/thing, that is creator of an audio
segment.
"""

from .issuer import Issuer  # noqa: F401

from .speaker import Speaker  # noqa: F401
from .speaker import Gender  # noqa: F401
from .speaker import AgeGroup  # noqa: F401

from .artist import Artist  # noqa: F401



import collections

from audiomate.utils import textfile


def write_file(path, entries):
    """
    Writes a ctm file.

    Args:
        path (str): Path to write the file to.
        entries (list): List with entries to write. (entries -> wave-file, channel, start (seconds),
                        duration (seconds), label)

    Example::

        >>> data = [
        >>>     ["wave-ab", '1', 0.0, 0.82, "duda"],
        >>>     ["wave-xy", '1', 0.82, 0.57, "Jacques"],
        >>> ]
        >>>
        >>> write_file('/path/to/file.txt', data)
    """

    textfile.write_separated_lines(path, entries, separator=' ')


def read_file(path):
    """
    Reads a ctm file.

    Args:
        path (str): Path to the file

    Returns:
        (dict): Dictionary with entries.

    Example::

        >>> read_file('/path/to/file.txt')
        {
            'wave-ab': [
                ['1', 0.00, 0.07, 'HI', 1],
                ['1', 0.09, 0.08, 'AH', 1]
            ],
            'wave-xy': [
                ['1', 0.00, 0.07, 'HI', 1],
                ['1', 0.09, 0.08, 'AH', 1]
            ]
        }
    """
    gen = textfile.read_separated_lines_generator(path, max_columns=6,
                                                  ignore_lines_starting_with=[';;'])

    utterances = collections.defaultdict(list)

    for record in gen:
        values = record[1:len(record)]

        for i in range(len(values)):
            if i == 1 or i == 2 or i == 4:
                values[i] = float(values[i])

        utterances[record[0]].append(values)

    return utterances
